
# file_prefix = 'C:/Users/rringuet/Kamodo_WinDev1/GITM/3DLST_t150317'
# varnames in cdf files are standardized (value[0])
model_varnames = {'r_Ar': ['mmr_Ar', 'mass mixing ratio of argon/neutrals',
                           0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           ''],
                  'rho_Ar': ['rho_Ar', 'mass density of argon',
                             1, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                             'kg/m**3'],
                  'r_CH4': ['mmr_CH4', 'mass mixing ratio of methane/neutrals',
                            2, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                            ''],
                  'k': ['k', 'total conduction',
                        3, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                        'W/m/K'],
                  'Q_EUV': ['Q_EUV', 'EUV heating',
                            4, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                            'K per timestep'],
                  'rho_H': ['rho_H', 'mass density of hydrogen',
                            5, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                            'kg/m**3'],
                  'rho_Hplus': ['rho_Hplus', 'mass density of hydrogen ion',
                                6, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                  'height'], 'kg/m**3'],
                  'r_H2': ['mmr_H2', 'mass mixing ratio of molecular ' +
                           'hydrogen/neutrals', 7, 'GDZ', 'sph',
                           ['time', 'lon', 'lat', 'height'], ''],
                  'r_HCN': ['mmr_HCN', 'mass mixing ratio of hydrogen ' +
                            'cyanide/neutrals', 8, 'GDZ', 'sph',
                            ['time', 'lon', 'lat', 'height'], ''],
                  'rho_He': ['rho_He', 'mass density of helium',
                             9, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                             'kg/m**3'],
                  'rho_Heplus': ['rho_Heplus', 'mass density of helium ion',
                                 10, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                    'height'], 'kg/m**3'],
                  'HeatingEfficiency': ['Q_eff', 'heating efficiency', 11,
                                        'GDZ', 'sph', ['time', 'lon', 'lat',
                                                       'height'], ''],
                  'HeatBalanceTotall': ['Q_bal', 'heat balance total', 12,
                                        'GDZ', 'sph', ['time', 'lon', 'lat',
                                                       'height'], ''],
                  'rho_N2': ['rho_N2', 'mass density of molecular nitrogen',
                             13, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                'height'], 'kg/m**3'],
                  'rho_N2plus': ['rho_N2plus', 'mass density of molecular ' +
                                 'nitrogen ion', 14, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'rho_Nplus': ['rho_Nplus', 'mass density of atomic ' +
                                'nitrogen ion', 15, 'GDZ', 'sph',
                                ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'rho_N2D': ['rho_N2D', 'mass density of atomic nitrogen ' +
                              '(2D state)', 16, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'rho_N2P': ['rho_N2P', 'mass density of atomic nitrogen ' +
                              '(2P state)', 17, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'rho_N4S': ['rho_N4S', 'mass density of atomic nitrogen ' +
                              '(4S state)', 18, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'r_N2': ['mmr_N2', 'mass mixing ratio of molecular nitrogen',
                           19, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           ''],
                  'rho_NO': ['rho_NO', 'mass density of nitric oxide',
                             20, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                'height'], 'kg/m**3'],
                  'rho_NOplus': ['rho_NOplus', 'mass density of nitric oxide' +
                                 ' ion', 21, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'rho_O2': ['rho_O2', 'mass density of molecular oxygen', 22,
                             'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                             'kg/m**3'],
                  'rho_O1D': ['rho_O1D', 'mass density of atomic oxygen ' +
                              '(1D state)', 23, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'rho_O2plus': ['rho_O2plus', 'mass density of molecular ' +
                                 'oxygen ion', 24, 'GDZ', 'sph',
                                 ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'rho_O2D': ['rho_O2D', 'mass density of atomic oxygen ' +
                              '(2D state)', 25, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], '1/m**3'],
                  'rho_Oplus2P': ['rho_Oplus2P', 'mass density of atomic ' +
                                  'oxygen ion (2P state)', 26, 'GDZ', 'sph',
                                  ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'rho_O3P': ['rho_O3P', 'mass density of atomic oxygen ' +
                              '(3P state)', 27, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'rho_Oplus4SP': ['rho_Oplus4S4P', 'mass density of atomic ' +
                                   'oxygen ion (4S or 4P state)', 28, 'GDZ',
                                   'sph', ['time', 'lon', 'lat', 'height'],
                                   'kg/m**3'],
                  'L_Rad': ['Q_cool', 'radiative cooling rate', 29, 'GDZ',
                            'sph',  ['time', 'lon', 'lat', 'height'], ''],
                  'rho': ['rho_n', 'neutral mass density', 30, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'T_n': ['T_n', 'neutral temperature', 31, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'K'],
                  'vi_east': ['v_ieast', 'zonal ion wind velocity (east)',
                              32, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                 'height'], 'm/s'],
                  'vi_north': ['v_inorth', 'meridional ion wind velocity ' +
                               '(north)', 33, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'vi_up': ['v_iup', 'vertical ion wind velocity (up)',
                            34, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                               'height'], 'm/s'],
                  'vn_east': ['v_neast', 'zonal neutral wind velocity (east)',
                              35, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                 'height'], 'm/s'],
                  'vn_north': ['v_nnorth', 'meridional neutral wind velocity' +
                               ' (north)', 36, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'vn_up': ['v_nup', 'vertical neutral wind velocity (up)',
                            37, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                            'm/s'],
                  'v_N2_up': ['v_N2up', 'vertical velocity of molecular ' +
                              'nitrogen (up)', 38, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'v_N4S_up': ['v_Nstate4Sup', 'vertical velocity of atomic ' +
                               'nitrogen (4S state) (up)', 39, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'v_N_up': ['v_Nup', 'vertical velocity of atomic nitrogen ' +
                             '(up)', 40, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                        'height'], 'm/s'],
                  'v_O2_up': ['v_O2up', 'vertical velocity of molecular ' +
                              'oxygen (up)', 41, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'v_O3P_up': ['v_Ostate3Pup', 'vertical velocity of atomic' +
                               ' (3P state) (up)', 42, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'height'], 'm/s'],
                  'v_He_up': ['v_Heup', 'vertical velocity of helium (up)',
                              43, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                 'height'], 'm/s'],
                  'N_e': ['N_e', 'electron number density', 44, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], '1/m**3'],
                  'ElectronAverageEnergy': ['E_eavg', 'average electron ' +
                                            'energy', 45, 'GDZ', 'sph',
                                            ['time', 'lon', 'lat', 'height'],
                                            'J'],
                  'T_e': ['T_e', 'electron temperature', 46, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'K'],
                  'T_i': ['T_i', 'ion temperature', 47, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], 'K'],
                  'SolarZenithAngle': ['SZA', 'solar zenith angle', 48, 'GDZ',
                                       'sph',  ['time', 'lon', 'lat'],
                                       'radians'],
                  'rho_CO2': ['rho_CO2', 'mass density of carbon dioxide', 49,
                              'GDZ', 'sph',  ['time', 'lon', 'lat', 'height'],
                              'kg/m**3'],
                  'DivJu FL': ['DivI_nfl', 'divergence of the neutral ' +
                               'wind-driven currents integrated along the ' +
                               'field-line', 50, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'height'], ''],
                  'DivJuAlt': ['DivI_nalt', 'divergence of the neutral ' +
                               'wind-driven currents integrated along the ' +
                               'altitude', 51, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'height'], ''],
                  'ElectronEnergyFlux': ['Phi_eE', 'electron energy flux', 52,
                                         'GDZ', 'sph', ['time', 'lon', 'lat',
                                                        'height'], 'J/m**2'],
                  'Field Line Length': ['s_Bfield', 'magnetic field arc ' +
                                        'line length', 53, 'GDZ', 'sph',
                                        ['time', 'lon', 'lat', 'height'], 'm'],
                  'sigma_P': ['sigma_P', 'Pedersen conductivity', 54, 'GDZ',
                              'sph',  ['time', 'lon', 'lat', 'height'], 'S/m'],
                  'V': ['V', 'electric potential', 57, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'height'], 'V'],
                  'sigma_H': ['sigma_H', 'Hall conductivity', 58, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'S/m'],
                  'I_R2': ['j_R2', 'region 2 electric current density',
                           59, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           'A/m**2'],
                  'I_R1': ['j_R1', 'region 1 electric current density',
                           60, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           'A/m**2'],
                  'Ed1': ['E_perpeast', 'dynamo electric field in the' +
                          ' perpendicular to the magnetic field direction ' +
                          'that is "eastward"', 61, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], ''],
                  'Ed2': ['E_perpnorth', 'dynamo electric field in the ' +
                          'perpendicular to the magnetic field direction ' +
                          'that is "northward"', 62, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], ''],
                  'SolarLocalTime': ['SLT', 'solar local time', 63, 'GDZ',
                                     'sph', ['time', 'lon', 'lat'], 'hr'],
                  'E_up': ['E_up', 'vertical electric field velocity (up)',
                           64, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           'V/m'],
                  'E_east': ['E_east', 'zonal electric field (east)',
                             65, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                'height'], 'V/m'],
                  'E_north': ['E_north', 'meridional electric field (north)',
                              66, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                 'height'], 'V/m'],
                  'E_mag': ['E_mag', 'magnitude of electric field',
                            67, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                            'V/m'],
                  'B_up': ['B_up', 'vertical magnetic field velocity (up)',
                           68, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           'nT'],
                  'B_east': ['B_east', 'zonal magnetic field (east)',
                             69, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                'height'], 'nT'],
                  'B_north': ['B_north', 'meridional magnetic field (north)',
                              70, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                 'height'], 'nT'],
                  'B_mag': ['B_mag', 'magnitude of magnetic field', 71, 'GDZ',
                            'sph', ['time', 'lon', 'lat', 'height'], 'nT'],
                  'MagLat': ['lat_B', 'magnetic latitude', 72, 'GDZ', 'sph',
                             ['time', 'lon', 'lat', 'height'], 'deg'],
                  'MagLon': ['lon_B', 'magnetic longitude', 73, 'GDZ', 'sph',
                             ['time', 'lon', 'lat', 'height'], 'deg'],
                  'g': ['g', 'gravitational acceleration', 74, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'height'], 'm/s**2'],
                  'GradP_east': ['GradP_east', 'zonal component of gradient ' +
                                 'of sum of ion and electron pressures (east)',
                                 75, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                    'height'], 'Pa/m'],
                  'GradP_north': ['GradP_north', 'meridional component of ' +
                                  'gradient of sum of ion and electron ' +
                                  'pressures (north)', 76, 'GDZ', 'sph',
                                  ['time', 'lon', 'lat', 'height'], 'Pa/m'],
                  'GradP_up': ['GradP_up', 'vertical component of gradient ' +
                               'of sum of ion and electron pressures (up)',
                               77, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                  'height'], 'Pa/m'],
                  'nu_in': ['nu_ion', 'ion neutral collision frequency', 78,
                            'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                            '1/s'],
                  'ChemicalHeatingRate': ['Q_chem', 'chemical heating rate',
                                          79, 'GDZ', 'sph', ['time', 'lon',
                                                             'lat', 'height'],
                                          ''],
                  'TotalAbsoluteEUV': ['Q_EUVabs', 'total absolute EUV', 80,
                                       'GDZ', 'sph',  ['time', 'lon', 'lat',
                                                       'height'],
                                       'K per timestep'],
                  'Q_Ocool': ['Q_Ocool', 'cooling rate of atomic oxygen', 81,
                              'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                              'K per timestep'],
                  'Q_Joule': ['Q_Joule', 'joule heating', 82, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'],
                              'K per timestep'],
                  'Q_Auroral': ['Q_auroral', 'auroral heating', 83, 'GDZ',
                                'sph', ['time', 'lon', 'lat', 'height'],
                                'K per timestep'],
                  'Q_PhotoE': ['Q_photoe', 'heating due to the ' +
                               'photoelectric effect', 84, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'height'],
                               'K per timestep'],
                  'k_eddy': ['k_ed', 'eddy conduction', 85, 'GDZ', 'sph',
                             ['time', 'lon', 'lat', 'height'], ''],
                  'k_adiabaticeddy': ['k_edadiab', 'adiabatic eddy conduction',
                                      86, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                         'height'], ''],
                  'Q_NOcool': ['Q_NOcool', 'cooling rate of nitric oxide', 87,
                               'GDZ', 'sph',  ['time', 'lon', 'lat', 'height'],
                               'K per timestep'],
                  'k_molecular': ['k_mol', 'molecular conduction', 88, 'GDZ',
                                  'sph', ['time', 'lon', 'lat', 'height'], ''],
                  'NmF2': ['NmF2', 'Maximum electron number density in ' +
                           'F2 layer', 89, 'GDZ', 'sph',
                           ['time', 'lon', 'lat'], ''],
                  'hmF2': ['hmF2', 'Height of maximum electron number ' +
                           'density in F2 layer', 90, 'GDZ', 'sph',
                           ['time', 'lon', 'lat'], 'km'],
                  'TEC': ['TEC', 'vertical total electron content ' +
                          '(height integrated from bottom to top boundary)',
                          91, 'GDZ', 'sph', ['time', 'lon', 'lat'],
                          '10**16/m**2'],
                  'Phi_Joule': ['Phi_Joule', 'joule heat flux', 92, 'GDZ',
                                'sph', ['time', 'lon', 'lat'], 'W/m**2'],
                  'Phi_Q': ['Phi_heat', 'heat flux', 93, 'GDZ', 'sph',
                            ['time', 'lon', 'lat'], 'W/m**2'],
                  'Phi_EUV': ['Phi_EUV', 'EUV heat flux', 94, 'GDZ', 'sph',
                              ['time', 'lon', 'lat'], 'W/m**2'],
                  'Phi_NOCooling': ['Phi_NOCooling', 'NO cooling flux', 95,
                                    'GDZ', 'sph', ['time', 'lon', 'lat'],
                                    'W/m**2']
                  }


def MODEL():
    from time import perf_counter
    from glob import glob
    from os.path import basename, isfile
    from numpy import array, unique, NaN
    from datetime import datetime, timezone
    from netCDF4 import Dataset
    from kamodo import Kamodo
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''GITM model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - GITM output files are given in multiple binary files per time
              step. No attempt is made to combine these files into a single
              file per time step, but each is converted into a netCDF4 file
              using code adapted from a script written by Dan Welling and
              Angeline Burrell (used with permission).
            - The converted files are small and are created with one time step
              per file, so interpolation method 1 is chosen. The standard SciPy
              interpolator is used.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     filetime=False, verbose=False, gridded_int=True,
                     printfiles=False, **kwargs):
            super(MODEL, self).__init__()
            self.modelname = 'GITM'

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # determine if calc of 2D variables is necessary
                total_files = sorted(glob(file_dir+'*.nc'))
                if sum(['2D' in file for file in total_files]) > 0:
                    flag_2D = True  # 2D files found, will not calc 2D vars
                else:  # try bin files...
                    bin_files = sorted(glob(file_dir+'2D*.bin'))
                    if len(bin_files) > 0:
                        flag_2D = True
                    else:
                        flag_2D = False  # calc will occur

                # check for and convert any files not converted yet
                from kamodo_ccmc.readers.gitm_tocdf import GITMbin_toCDF as\
                    toCDF
                toCDF(file_dir, flag_2D)

                t0 = perf_counter()  # begin timer
                # figure out types of files present (2DTEC, 3DALL, 3DLST, etc)
                total_files = sorted(glob(file_dir+'*.nc'))
                patterns = sorted(unique([basename(f).split('_t')[0] for f in
                                          total_files]))
                self.filename = ''.join([f+',' for f in total_files])[:-1]

                # establish time attributes
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+p+'*.nc'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times
                    for f in range(len(pattern_files)):
                        cdf_data = Dataset(pattern_files[f])
                        # hours since midnight first file
                        tmp = array(cdf_data.variables['time'])
                        self.times[p]['start'].append(tmp[0])
                        self.times[p]['end'].append(tmp[-1])
                        self.times[p]['all'].extend(tmp)
                        cdf_data.close()
                    self.times[p]['start'] = array(self.times[p]['start'])
                    self.times[p]['end'] = array(self.times[p]['end'])
                    self.times[p]['all'] = array(self.times[p]['all'])

                # establish date of first file
                cdf_data = Dataset(total_files[0], 'r')
                string_date = cdf_data.filedate
                cdf_data.close()
                self.filedate = datetime.strptime(
                    string_date+' 00:00:00', '%Y-%m-%d %H:%M:%S'
                    ).replace(tzinfo=timezone.utc)

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times as is to prevent infinite recursion

            # store variables
            self.missing_value = NaN
            self.varfiles = {}  # store which variable came from which file
            self.gvarfiles = {}  # store file variable name similarly
            self.err_list = []
            self.variables = {}

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item not in
                            test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # loop through file patterns for coordinate grids + var mapping
            for p in self.pattern_files.keys():
                pattern_files = self.pattern_files[p]

                # get coordinates from first file
                cdf_data = Dataset(pattern_files[0], 'r')
                setattr(self, '_lat_'+p, array(cdf_data.variables['lat']))
                setattr(self, '_lon_'+p, array(cdf_data.variables['lon']))
                if '2D' not in p:
                    setattr(self, '_height_'+p,
                            array(cdf_data.variables['height']))

                # check var_list for variables not possible in this file set
                if len(variables_requested) > 0 and\
                        variables_requested != 'all':
                    gvar_list = [key for key in model_varnames.keys()
                                 if key in cdf_data.variables.keys() and
                                 model_varnames[key][0] in variables_requested]
                    if len(gvar_list) != len(variables_requested):
                        err_list = [value[0] for key, value in
                                    model_varnames.items()
                                    if key not in cdf_data.variables.keys() and
                                    value[0] in variables_requested]
                        self.err_list.extend(err_list)  # add to master list
                else:
                    gvar_list = [key for key in model_varnames.keys()
                                 if key in cdf_data.variables.keys()]
                # store which file these variables came from
                self.varfiles[p] = [model_varnames[key][0] for
                                    key in gvar_list]
                self.gvarfiles[p] = gvar_list
                cdf_data.close()

            # collect all possible variables in set of files and return
            var_list = []
            for p in self.varfiles.keys():
                var_list.extend(self.varfiles[p])
            if variables_requested == 'all':
                self.var_dict = {value[0]: value[1:] for key, value in
                                 model_varnames.items() if value[0] in
                                 var_list}
                return

            # loop through patterns to store variable keys and units
            for p in self.pattern_files.keys():
                # initialize variable dictionaries
                for gvar in self.gvarfiles[p]:
                    self.variables[model_varnames[gvar][0]] = {
                        'units': model_varnames[gvar][-1], 'data': p}

            # remove successful variables from err_list
            self.err_list = list(unique(self.err_list))
            self.err_list = [item for item in self.err_list if item not in
                             self.variables.keys()]
            if len(self.err_list) > 0:
                print('Some requested variables are not available in the ' +
                      'files found:\n',
                      self.pattern_files.keys(), self.err_list)

            if printfiles:
                print(f'{len(self.filename)} Files:')
                files = self.filename.split(',')
                for f in files:
                    print(f)

            # return if only one time found - interpolator code will break
            if ',' not in self.filename:
                print('Not enough times found in given directory.')
                return

            # register interpolators for each variable
            t_reg = perf_counter()
            # store original list b/c gridded interpolators change keys list
            varname_list = [key for key in self.variables.keys()]
            for varname in varname_list:
                self.register_variable(varname, gridded_int)

            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        def register_variable(self, varname, gridded_int):
            """Registers an interpolator with proper signature"""

            # determine which file the variable came from, retrieve the coords
            key = self.variables[varname]['data']
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]  # variable name in file
            coord_list = [value[-2] for key, value in
                          model_varnames.items() if value[0] == varname][0]
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[key]['all']}}
            # get the correct coordinates
            coord_dict['lon'] = {
                'data': getattr(self, '_lon_'+key), 'units': 'deg'}
            coord_dict['lat'] = {
                'data': getattr(self, '_lat_'+key), 'units': 'deg'}
            if len(coord_list) == 4:
                coord_dict['height'] = {
                    'data': getattr(self, '_height_'+key),
                    'units': 'km'}
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]

            def func(i):
                '''i is the file number.'''
                # get data from file
                file = self.pattern_files[key][i]
                cdf_data = Dataset(file)
                data = array(cdf_data.variables[gvar])
                cdf_data.close()
                return data

            # define and register the interpolators
            self = RU.Functionalize_Dataset(self, coord_dict, varname,
                                            self.variables[varname],
                                            gridded_int, coord_str,
                                            interp_flag=1, func=func)
    return MODEL



import kamodo # get installed kamodo

# insert Kamodo class into kamodo namespace
from kamodo.kamodo import Kamodo
setattr(kamodo, 'Kamodo', Kamodo)

# make this directory's files accessible
import os, sys; sys.path.append(os.path.dirname(os.path.realpath(__file__)))
import warnings




'''
TIEGCM Kamodo reader, adapted to new structure for satellite flythrough
    software
Initial version - Asher Pembroke (?)
Initial version of model_varnames contributed by Zachary Waldron
New code: Rebecca Ringuette (June 2021 and on)

The high-altitude version of TIEGCM has one timestep per file, while the normal
version of TIEGCM has multiple timesteps per file. This difference requires
different interp_flag values and different functions. Adding the logic for
this at the end. - Nov 2, 2022

NOTE:
    The current logic for variables that depend on imlev slices off e36 values
        in self._imlev coordinate array. This only works because there is one
        variable that depends on imlev: H_imlev. The logic on lines 311-313
        will have to be reworked a bit if other variables depend on imlev
        later.
'''
from numpy import vectorize
from datetime import datetime, timezone, timedelta

model_varnames = {  # 4D Variables, vert coordinate on midpoint levels (ilev)
                  # geometric height- interpolated to the mid points
                  "ZGMID": ["H_ilev",
                            'height dependent on primary pressure level',
                            0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                            "cm"],
                  "TN": ["T_n_ilev", 'neutral temperature',
                         1, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'], "K"],
                  "TN2": ["T_n", 'neutral temperature', 1, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], "K"],
                  "O2": ["mmr_O2_ilev",
                         'mass mixing ratio of molecular oxygen', 2, 'GDZ',
                         'sph', ['time', 'lon', 'lat', 'ilev'], ""],
                  "O22": ["mmr_O2", 'mass mixing ratio of molecular oxygen',
                          2, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                          ""],
                  "O1": ["mmr_O_ilev", 'mass mixing ratio of atomic oxygen',
                         3, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'], ""],
                  "O12": ["mmr_O", 'mass mixing ratio of atomic oxygen', 3,
                          'GDZ', 'sph', ['time', 'lon', 'lat', 'height'], ""],
                  "N2": ["mmr_N2_ilev",
                         'mass mixing ratio of molecular nitrogen', 4, 'GDZ',
                         'sph', ['time', 'lon', 'lat', 'ilev'], ""],
                  "N22": ["mmr_N2", 'mass mixing ratio of molecular nitrogen',
                          4, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                          ""],
                  "HE": ["mmr_He_ilev", 'mass mixing ratio of atomic helium',
                         5, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'], ""],
                  "HE2": ["mmr_He", 'mass mixing ratio of atomic helium', 5,
                          'GDZ', 'sph', ['time', 'lon', 'lat', 'height'], ""],
                  "NO": ["mmr_NO_ilev",
                         'mass mixing ratio of molecular nitric oxide', 6,
                         'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'], ""],
                  "NO2": ["mmr_NO",
                          'mass mixing ratio of molecular nitric oxide', 6,
                          'GDZ', 'sph', ['time', 'lon', 'lat', 'height'], ""],
                  "N4S": ["mmr_Nstate4S_ilev",
                          'mass mixing ratio of atomic nitrogen (4S state)',
                          7, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'], ""],
                  "N4S2": ["mmr_Nstate4S",
                           'mass mixing ratio of atomic nitrogen (4S state)',
                           7, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           ""],
                  "N2D": ["mmr_Nstate2D_ilev",
                          'mass mixing ratio of atomic nitrogen (2D state)',
                          8, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'], ""],
                  "N2D2": ["mmr_Nstate2D",
                           'mass mixing ratio of atomic nitrogen (2D state)',
                           8, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           ""],
                  "TE": ["T_e_ilev", 'electron temperature', 9, 'GDZ', 'sph',
                         ['time', 'lon', 'lat', 'ilev'], "K"],
                  "TE2": ["T_e", 'electron temperature', 9, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], "K"],
                  "TI": ["T_i_ilev", 'ion temperature', 10, 'GDZ', 'sph',
                         ['time', 'lon', 'lat', 'ilev'], "K"],
                  "TI2": ["T_i", 'ion temperature', 10, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], "K"],
                  "O2P": ["N_O2plus_ilev",
                          'number density of molecular oxygen ion', 11, 'GDZ',
                          'sph', ['time', 'lon', 'lat', 'ilev'], "1/cm**3"],
                  "O2P2": ["N_O2plus",
                           'number density of molecular oxygen ion', 11, 'GDZ',
                           'sph', ['time', 'lon', 'lat', 'height'], "1/cm**3"],
                  "OP": ["N_Oplus_ilev", 'number density of atomic oxygen ion',
                         12, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                         "1/cm**3"],
                  "OP2": ["N_Oplus", 'number density of atomic oxygen ion',
                          12, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                          "1/cm**3"],
                  "N2N": ["N_N2_ilev", 'number density of molecular nitrogen',
                          13, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                          "1/cm**3"],
                  "N2N2": ["N_N2", 'number density of molecular nitrogen',
                           13, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           "1/cm**3"],
                  "CO2_COOL": ["Q_CO2cool_ilev",
                               'cooling rate of carbon dioxide', 14, 'GDZ',
                               'sph', ['time', 'lon', 'lat', 'ilev'],
                               "erg/g/s"],
                  "CO2_COOL2": ["Q_CO2cool", 'cooling rate of carbon dioxide',
                                14, 'GDZ', 'sph',
                                ['time', 'lon', 'lat', 'height'], "erg/g/s"],
                  "NO_COOL": ["Q_NOcool_ilev", 'cooling rate of nitric oxide',
                              15, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                              "erg/g/s"],
                  "NO_COOL2": ["Q_NOcool", 'cooling rate of nitric oxide',
                               15, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                  'height'], "erg/g/s"],
                  "UN": ["v_neast_ilev", 'zonal neutral wind velocity (east)',
                         16, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                         "cm/s"],
                  "UN2": ["v_neast", 'zonal neutral wind velocity (east)',
                          16, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                          "cm/s"],
                  "VN": ["v_nnorth_ilev",
                         'meridional neutral wind velocity (north)', 17, 'GDZ',
                         'sph', ['time', 'lon', 'lat', 'ilev'], "cm/s"],
                  "VN2": ["v_nnorth",
                          'meridional neutral wind velocity (north)', 17,
                          'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                          "cm/s"],
                  # "O2P_ELD": ['O2P_ELD', '???',18,'GDZ', 'sph',
                  # ['time', 'lon', 'lat', 'ilev'], ''], #NO DESCRIPTION GIVEN
                  # "N2P_ELD": ['N2P_ELD', '???',19,'GDZ', 'sph',
                  # ['time', 'lon', 'lat', 'ilev'], ''], #NO DESCRIPTION GIVEN
                  "NPLUS": ['N_Nplus_ilev',
                            'number density of atomic nitrogen ion', 20, 'GDZ',
                            'sph', ['time', 'lon', 'lat', 'ilev'], '1/cm**3'],
                  "NPLUS2": ['N_Nplus',
                             'number density of atomic nitrogen ion', 20,
                             'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                             '1/cm**3'],
                  # "NOP_ELD": ['NOP_ELD', '???',21,'GDZ', 'sph',
                  # ['time', 'lon', 'lat', 'ilev'], ''], #NO DESCRIPTION GIVEN
                  "SIGMA_PED": ['sigma_P_ilev', 'Pedersen conductivity', 22,
                                'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                                'S/m'],
                  "SIGMA_PED2": ['sigma_P', 'Pedersen conductivity', 22, 'GDZ',
                                 'sph', ['time', 'lon', 'lat', 'height'],
                                 'S/m'],
                  "SIGMA_HAL": ['sigma_H_ilev', 'Hall conductivity', 23, 'GDZ',
                                'sph', ['time', 'lon', 'lat', 'ilev'], 'S/m'],
                  "SIGMA_HAL2": ['sigma_H', 'Hall conductivity', 23, 'GDZ',
                                 'sph', ['time', 'lon', 'lat', 'height'],
                                 'S/m'],
                  "QJOULE": ['Q_Joule_ilev', 'joule heating', 24, 'GDZ', 'sph',
                             ['time', 'lon', 'lat', 'ilev'], 'erg/g/s'],
                  "QJOULE2": ['Q_Joule', 'joule heating', 24, 'GDZ', 'sph',
                              ['time', 'lon', 'lat', 'height'], 'erg/g/s'],
                  "O_N2": ['OtoN2_ilev', 'Oxygen/molecular nitrogen ratio', 25,
                           'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'], ''],
                  "O_N22": ['OtoN2', 'Oxygen/molecular nitrogen ratio', 25,
                            'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                            ''],
                  # "N2D_ELD": ['N2D_ELD', '???',26,'GDZ', 'sph',
                  # ['time', 'lon', 'lat', 'ilev'], ''],  #NO DESCRIPTION GIVEN
                  "O2N": ['N_O2_ilev', 'number density of molecular oxygen',
                          27, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],
                          '1/cm**3'],
                  "O2N2": ['N_O2', 'number density of molecular oxygen', 27,
                           'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                           '1/cm**3'],

                  # 4D Variables, vert coordinate on interface levels (ilev1)
                  "DEN": ["rho_ilev1", 'total mass density', 28, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'ilev1'], "g/cm**3"],
                  "DEN2": ["rho", 'total mass density', 28, 'GDZ', 'sph',
                           ['time', 'lon', 'lat', 'height'], "g/cm**3"],
                  "ZG": ["H_ilev1",
                         'height dependent on secondary pressure level', 29,
                         'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev1'], "cm"],
                  "Z": ["H_geopot", 'geopotential height', 30, 'GDZ', 'sph',
                        ['time', 'lon', 'lat', 'ilev1'], "cm"],
                  "NE": ["N_e_ilev", 'electron number density', 31, 'GDZ',
                         'sph', ['time', 'lon', 'lat', 'ilev1'], "1/cm**3"],
                  "NE2": ["N_e", 'electron number density', 31, 'GDZ', 'sph',
                          ['time', 'lon', 'lat', 'height'], "1/cm**3"],
                  "OMEGA": ["omega_ilev", 'Vertical motion frequency', 32,
                            'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev1'],
                            "1/s"],
                  "OMEGA2": ["omega", 'Vertical motion frequency', 32, 'GDZ',
                             'sph', ['time', 'lon', 'lat', 'height'], "1/s"],
                  "POTEN": ["V_ilev", 'electric potential', 33, 'GDZ', 'sph',
                            ['time', 'lon', 'lat', 'ilev1'], "V"],
                  "POTEN2": ["V", 'electric potential', 33, 'GDZ', 'sph',
                             ['time', 'lon', 'lat', 'height'], "V"],
                  "UI_ExB": ["v_iExBeast_ilev",
                             'zonal ExB ion velocity (east)', 34, 'GDZ', 'sph',
                             ['time', 'lon', 'lat', 'ilev1'], 'cm/s'],
                  "UI_ExB2": ["v_iExBeast", 'zonal ExB ion velocity (east)',
                              34, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                 'height'], 'cm/s'],
                  "VI_ExB": ["v_iExBnorth_ilev",
                             'meridional ExB ion velocity (north)', 35, 'GDZ',
                             'sph', ['time', 'lon', 'lat', 'ilev1'], 'cm/s'],
                  "VI_ExB2": ["v_iExBnorth",
                              'meridional ExB ion velocity (north)', 35, 'GDZ',
                              'sph', ['time', 'lon', 'lat', 'height'], 'cm/s'],
                  "WI_ExB": ["v_iExBup_ilev", 'vertical ExB ion velocity (up)',
                             36, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev1'],
                             'cm/s'],
                  "WI_ExB2": ["v_iExBup", 'vertical ExB ion velocity (up)',
                              36, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                 'height'], 'cm/s'],

                  # 4D Variables, vert coord on interface mag levels (imlev)
                  "ZMAG": ["H_milev",
                           'height dependent on geomagnetic pressure level',
                           37, 'MAG', 'sph', ['time', 'mlon', 'mlat', 'milev'],
                           "km"],

                  # 3D Variables,    (time, lat, lon)
                  "TEC": ["TEC", 'vertical total electron content (height ' +
                          'integrated from bottom to top boundary)',
                          38, 'GDZ', 'sph', ['time', 'lon', 'lat'], "1/cm**2"],
                  # Lower boundary condition for TN
                  # "TLBC": ["T_nLBC", 'Lower boundary condition for T_n',39,
                  # 'GDZ', 'sph', ['time', 'lon', 'lat'], "K"],
                  # Lower boundary condition for UN
                  # "ULBC": ["v_neastLBC", 'Lower boundary condition for v_n
                  # east component',40,'GDZ', 'sph', ['time', 'lon', 'lat'],
                  # "cm/s"],
                  # Lower boundary condition for VN
                  # "VLBC": ["v_nnorthLBC", 'Lower boundary condition for v_n
                  # north component',41,'GDZ', 'sph', ['time', 'lon', 'lat'],
                  # "cm/s"],
                  # Lower boundary condition for TN (TIME N-1)
                  # "TLBC_NM": ["T_nLBCNminus1", 'Lower boundary condition for
                  # T_n at t=N-1',42,'GDZ', 'sph', ['time', 'lon', 'lat'],
                  # "K"],
                  # Lower boundary condition for UN (TIME N-1)
                  # "ULBC_NM": ["v_neastLBCNminus1", 'Lower boundary condition
                  # for v_n north component at t=N-1',43,'GDZ', 'sph', ['time',
                  # 'lon', 'lat'], "cm/s"],
                  # Lower boundary condition for VN (TIME N-1)
                  # "VLBC_NM": ["v_nnorthLBCNminus1", 'Lower boundary condition
                  # for v_n east component at t=N-1',44,'GDZ', 'sph', ['time',
                  # 'lon', 'lat'], "cm/s"],
                  "QJOULE_INTEG": ["W_JouleH",
                                   'height integrated joule heating', 45,
                                   'GDZ', 'sph', ['time', 'lon', 'lat'],
                                   'erg/cm**2/s'],
                  "EFLUX": ['Phi_E', 'energy flux', 46, 'GDZ', 'sph',
                            ['time', 'lon', 'lat'], 'erg/cm**2/s'],
                  "HMF2": ['HmF2', 'height of maximum electron number ' +
                           'density in F2 layer', 47, 'GDZ', 'sph',
                           ['time', 'lon', 'lat'], 'km'],
                  "NMF2": ['NmF2', 'maximum electron number density in F2 ' +
                           'layer', 48, 'GDZ', 'sph', ['time', 'lon', 'lat'],
                           '1/cm**3'],
                  }


def year_mtime_todt0(year, mtime):  # self.filedate
    '''Convert year and day to datetime object in UTC at midnight'''

    day, hour, minute = mtime  # unpack mtime values
    return datetime(int(year), 1, 1).replace(tzinfo=timezone.utc) +\
        timedelta(days=int(day-1))


def year_mtime_todt(year, mtime):
    '''Convert year and [day,hour,minute] to datetime object in UTC'''

    day, hour, minute = mtime  # unpack mtime values
    return datetime(int(year), 1, 1).replace(tzinfo=timezone.utc) +\
        timedelta(days=int(day-1), hours=int(hour), minutes=int(minute))


@vectorize
def year_mtime_tohrs(year, day, hour, minute, filedate):
    '''Convert year and mtime to hours since midnight using predetermined
    datetime object.'''

    mtime = [day, hour, minute]
    return (year_mtime_todt(year, mtime) - filedate).total_seconds()/3600.


def MODEL():
    from time import perf_counter
    from os.path import basename, isfile
    from glob import glob
    from numpy import zeros, transpose, array, append, insert, where, unique
    from numpy import NaN, mean, broadcast_to, cos, sin, sum, squeeze
    from numpy import pi as nppi
    from netCDF4 import Dataset
    from kamodo import Kamodo
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''TIEGCM model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - TIE-GCM model outputs are given in netCDF files, so no file
              conversion is needed. However, all of the variables depend on a
              pressure level coordinate, so some pre-processing is needed to
              calculate the median height across the entire dataset
              (time, lon, lat) for a given pressure level value for each type
              of pressure level coordinate (up to three).
            - TIE-GCM data is given in several coordinate systems depending
              on pressure level - a unitless representation of height
              corresponding to a given atmospheric pressure. The preset values
              of pressure level in the coordinate systems are not
              guaranteed to be identical, so they are inverted independently of
              the other unless only one is provided. In that case, the are
              assumed to be identical. The magnetic pressure level coordinate
              is always treated independently of the other two.
            - Pressure level inversion is performed in the reader_utilities
              script, specifically the PLevelInterp function. Two versions of
              all variables that depend on pressure level are created: the
              original and one dependent on height, which is created through
              Kamodo's function composition feature.
            - The outputs do not provide values at the poles, so scalar and
              vector averaging are used as appropriate to determine these
              values on the fly.
            - The files are small, but different versions require different
              interpolation choices. The high-altitude version of TIEGCM has
              one timestep per file, requiring that interpolation methods 1 be
              chosen, while the normal version of TIEGCM has multiple timesteps
              per file, requiring that interpolation method 2 be chosen. This
              difference requires custom code to automatically choose the
              appropriate interpolation method and the associated read logic.

        Developer note:
            The current logic for variables that depend on imlev slices off the
            e36 values in self._imlev coordinate array. This only works because
            there is only one variable that depends on imlev: H_imlev. The
            logic beginning on line 767 will have to be reworked a bit if any
            new variables depend on imlev to avoid repeatedly slicing off the
            coordinate values.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     filetime=False, verbose=False, gridded_int=True,
                     printfiles=False, **kwargs):
            super(MODEL, self).__init__()
            self.modelname = 'TIEGCM'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # collect filenames, all one pattern, so doesn't matter
                raw_files = sorted(glob(file_dir+'*.nc'))
                files = [f for f in raw_files if self.modelname not in
                         basename(f)]
                patterns = unique([basename(f)[0] for f in files])  # one patte
                self.filename = ''.join([f+',' for f in files])[:-1]

                # establish time attributes
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+p+'*.nc'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times
                    for f in range(len(pattern_files)):
                        cdf_data = Dataset(pattern_files[f])
                        year = array(cdf_data.variables['year'])
                        mtime = array(cdf_data.variables['mtime'])
                        day, hour, minute = mtime.T
                        # datetime object for file date at midnight UTC
                        if f == 0 and p == patterns[0]:
                            self.filedate = year_mtime_todt0(year[0], mtime[0])
                        time = year_mtime_tohrs(year, day, hour, minute,
                                                self.filedate)
                        cdf_data.close()
                        self.times[p]['start'].append(time[0])
                        self.times[p]['end'].append(time[-1])
                        self.times[p]['all'].extend(time)
                    self.times[p]['start'] = array(self.times[p]['start'])
                    self.times[p]['end'] = array(self.times[p]['end'])
                    self.times[p]['all'] = array(self.times[p]['all'])

                # perform post-processing to speed up pressure level conversion
                from kamodo_ccmc.readers.tiegcm_tocdf import convert_all
                convert_all(file_dir, self.pattern_files, self.times)

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times as is to prevent recursion

            # These lists are the standardized variable name.
            # The only milev variable is H_milev.
            ilev1_list = [value[0] for key, value in model_varnames.items()
                          if value[5][-1] == 'ilev1']
            ilev1_replace = [item.split('_ilev1')[0] for item in ilev1_list if
                             item != 'H_ilev1']
            ilev_list = [value[0] for key, value in model_varnames.items()
                         if value[5][-1] == 'ilev']
            ilev_replace = [item.split('_ilev')[0] for item in ilev_list if
                            item != 'H_ilev']
            # milev_list = [value[0] for key, value in model_varnames.items()
            #               if value[5][-1] == 'milev']
            self.total_ilev = [item for item in ilev_list + ilev1_list if item
                               not in ['H_ilev', 'H_ilev1']]
            self.total_replace = ilev_replace + ilev1_replace
            # dictionary mapping to navigate related variable names
            self.ilev_map = {item1: item2 for item1, item2 in
                             zip(self.total_replace, self.total_ilev)}

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item not in
                            test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # translate from standardized variables to names in file
            # remove variables requested that are not in the file
            p = list(self.pattern_files.keys())[0]
            pattern_files = self.pattern_files[p]
            cdf_data = Dataset(pattern_files[0])
            if len(variables_requested) > 0 and variables_requested != 'all':
                # add ilev version of variables to the list, adding H_ilev(1)
                add_ilev = [var+'_ilev' for var in variables_requested if var
                            in ilev_replace]
                add_ilev1 = [var+'_ilev1' for var in variables_requested if var
                             in ilev1_replace]
                if len(add_ilev) > 0:
                    add_ilev += ['H_ilev']
                if len(add_ilev1) > 0:
                    add_ilev1 += ['H_ilev1']
                new_var = variables_requested + add_ilev + add_ilev1
                short_var = [item for item in new_var if item not
                             in self.ilev_map.keys()]  # remove replaced items
                gvar_list = [key for key, value in model_varnames.items()
                             if value[0] in short_var and key in
                             cdf_data.variables.keys()]  # file variable names

                # check for variables requested but not available
                if len(gvar_list) != len(short_var):
                    err_list = [value[0] for key, value in
                                model_varnames.items() if value[0] in
                                short_var and key not in
                                cdf_data.variables.keys()]
                    if len(err_list) > 0:
                        print('Some requested variables are not available: ',
                              err_list)
                    if 'H_ilev' in err_list or 'H_ilev1' in err_list:
                        other_name = ['H_ilev', 'H_ilev1']
                        err_name = [name for name in other_name if name in
                                    err_list][0]
                        other_name.remove(err_name)  # see first element
                        gvar_name = [key for key, value in
                                     model_varnames.items() if value[0] ==
                                     other_name[0]]
                        gvar_list += gvar_name  # add other height name to list
                        print(f'Retrieving the {other_name[0]} variable' +
                              ' instead.')

            else:  # only input variables on the avoid_list if requested
                avoid_list = ['TLBC', 'ULBC', 'VLBC', 'TLBC_NM', 'ULBC_NM',
                              'VLBC_NM', 'NOP_ELD', 'O2P_ELD', 'N2P_ELD',
                              'N2D_ELD']
                gvar_list = [key for key in cdf_data.variables.keys()
                             if key in model_varnames.keys() and
                             key not in avoid_list]
                if variables_requested == 'all':
                    self.var_dict = {value[0]: value[1:] for key, value in
                                     model_varnames.items() if key in
                                     gvar_list}
                    # add non-ilev versions of the variables in the files
                    key_list = list(self.var_dict.keys())
                    for var_key in key_list:
                        if var_key in self.total_ilev:
                            # retrieve equivalent non-ilev variable name
                            new_key = [key for key, value in
                                       self.ilev_map.items() if value ==
                                       var_key][0]
                            # retrieve key for model_varnames mapping
                            model_key = [key for key, value in
                                         model_varnames.items() if
                                         value[0] == new_key][0]
                            # add to returned dictionary
                            self.var_dict[model_varnames[model_key][0]] =\
                                model_varnames[model_key][1:]

                    cdf_data.close()
                    return

            # Store the requested variables into a dictionary
            self.variables = {model_varnames[key][0]: {
                'units': model_varnames[key][-1],
                'data': key} for key in gvar_list}

            # Store inputs as class attributes
            self.missing_value = NaN
            self._registered = 0
            if printfiles:
                print('Files:', self.filename)

            # store coordinates
            lat = array(cdf_data.variables['lat'])  # NOT FULL RANGE IN LAT
            lat = insert(lat, 0, -90)  # insert a grid point before -87.5
            self._lat = append(lat, 90.)   # and at the end (after 87.5)
            lon = array(cdf_data.variables['lon'])  # NOT WRAPPED IN LONGITUDE
            self._lon = append(lon, 180.)  # add 180. to end of array
            self._ilev = array(cdf_data.variables['lev'])
            self._ilev1 = array(cdf_data.variables['ilev'])
            self._milev = array(cdf_data.variables['imlev'])
            self._mlat = array(cdf_data.variables['mlat'])
            self._mlon = array(cdf_data.variables['mlon'])  # -180 to 180
            if verbose:
                print(f'Took {perf_counter()-t0:.6f}s to read in data')

            # Check for presence of necessary height variables in varname_list.
            varname_list = [key for key in self.variables.keys()]
            ilev1_check = unique([True for item in varname_list if 'ilev1' ==
                                  item[-5:]])
            ilev_check = unique([True for item in varname_list if 'ilev' ==
                                 item[-4:]])
            if ilev1_check and 'H_ilev1' not in varname_list:
                self.ilev_sub = 'H_ilev1'  # name of H missing
            elif ilev_check and 'H_ilev' not in varname_list:
                self.ilev_sub = 'H_ilev'
            else:
                self.ilev_sub = False

            # register interpolators for each requested variable
            # rearrange to deal with H_ilev and H_ilev1 first if there
            # also calcalate median km grids
            if 'H_ilev' in varname_list:  # height in cm
                varname_list.remove('H_ilev')
                varname_list = ['H_ilev'] + varname_list
                if isfile(file_dir+'TIEGCM_km.nc'):  # km_ilev from file
                    km_data = Dataset(file_dir+'TIEGCM_km.nc')
                    if hasattr(km_data, 'km_ilev_max'):
                        self._km_ilev = array(km_data.variables['km_ilev'])
                        self._km_ilev_max = km_data.km_ilev_max
                        self._km_ilev_min = km_data.km_ilev_min
                    km_data.close()
            if 'H_ilev1' in varname_list:  # height in cm
                varname_list.remove('H_ilev1')
                varname_list = ['H_ilev1'] + varname_list
                if isfile(file_dir+'TIEGCM_km.nc'):  # km_ilev1 from file
                    km_data = Dataset(file_dir+'TIEGCM_km.nc')
                    if hasattr(km_data, 'km_ilev1_max'):
                        self._km_ilev1 = array(km_data.variables['km_ilev1'])
                        self._km_ilev1_max = km_data.km_ilev1_max
                        self._km_ilev1_min = km_data.km_ilev1_min
                    km_data.close()
            cdf_data.close()
            t_reg = perf_counter()
            for varname in varname_list:
                self.register_variable(varname, gridded_int, verbose)
            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        def wrap_3Dlatlon(self, varname, variable):
            '''Wraps the data array in longitude (-180=180), and latitude'''

            shape_list = list(variable.shape)
            shape_list[2] += 2  # need two more places in latitude
            shape_list[1] += 1  # need one more place in longitude
            tmp_arr = zeros(shape_list)  # array to set-up wrapped data in
            tmp_arr[0:, :-1, 1:-1] = variable  # copy data into grid

            # wrapping in latitude for scalar variables
            # put in top values
            top = mean(tmp_arr[:, :-1, 1], axis=1)  # same shape as time axis
            tmp_arr[:, :-1, 0] = broadcast_to(top, (shape_list[1]-1,
                                                    shape_list[0])).T
            # same for bottom, reusing variable names
            top = mean(tmp_arr[:, :-1, -2], axis=1)  # same shape as time axis
            tmp_arr[:, :-1, -1] = broadcast_to(top, (shape_list[1]-1,
                                                     shape_list[0])).T

            # wrap in longitude after to prevent double counting in average
            tmp_arr[:, -1, :] = tmp_arr[:, 0, :]
            return tmp_arr

        def wrap_4Dlatlon(self, varname, variable):
            '''Wraps the data array in longitude (-180=180), and latitude
            (0=-2, -1=1)'''

            shape_list = list(variable.shape)  # time, lon, lat, ilev
            shape_list[2] += 2  # need two more places in latitude
            shape_list[1] += 1  # need one more place in longitude
            tmp_arr = zeros(shape_list)  # array to set-up wrapped data in
            tmp_arr[:, :-1, 1:-1, :] = variable  # copy data into grid

            # wrapping in latitude for scalar and cartesian/radial variables
            if varname not in ['u_n', 'v_n', 'u_iExB', 'v_iExB']:
                # put in top values
                top = mean(tmp_arr[:, :-1, 1, :], axis=1)  # average over lon
                tmp_arr[:, :-1, 0, :] = transpose(broadcast_to(
                    top, (shape_list[1]-1, shape_list[0], shape_list[3])),
                    (1, 0, 2))
                # same for bottom, reusing variable names
                top = mean(tmp_arr[:, :-1, -2, :], axis=1)  # average over lon
                tmp_arr[:, :-1, -1, :] = transpose(broadcast_to(
                    top, (shape_list[1]-1, shape_list[0], shape_list[3])),
                    (1, 0, 2))
            # wrapping in latitude for relevant vector variables
            elif varname in ['u_n', 'v_n', 'u_iExB', 'v_iExB']:
                # calculate net vector magnitude for top
                tmp_arr[:, :-1, 0, :] = self.vector_average4D(
                    tmp_arr[:, :-1, 1, :], shape_list, varname, self._lat[0])
                # repeat for bottom
                tmp_arr[:, :-1, -1, :] = self.vector_average4D(
                    tmp_arr[:, :-1, -2, :], shape_list, varname, self._lat[-1])
            tmp_arr[:, -1, :, :] = tmp_arr[:, 0, :, :]  # wrap value in lon
            return tmp_arr

        def vector_average4D(self, top, shape_list, varname, latval):
            '''find vector average at pole for array with shape (time, lon,
            height)'''

            # find net x and y components, final array shapes are (t, lon, ht)
            lon_arr = transpose(broadcast_to(self._lon[:-1],
                                             (shape_list[0], shape_list[3],
                                              shape_list[1]-1)), (0, 2, 1))
            # need to put 'old' shape at end in broadcast_to call ^
            xval = sum(top*cos((lon_arr+180.)*nppi/180.), axis=1)  # same shape
            yval = sum(top*sin((lon_arr+180.)*nppi/180.), axis=1)  # as t and z
            xarr = transpose(broadcast_to(xval, (shape_list[1]-1,
                                                 shape_list[0],
                                                 shape_list[3])), (1, 0, 2))
            yarr = transpose(broadcast_to(yval, (shape_list[1]-1,
                                                 shape_list[0],
                                                 shape_list[3])), (1, 0, 2))

            # convert to proper unit vector (see wiki on spherical coordinates)
            if 'u' in varname:
                # Zonal/east components -> convert to psi_hat vector (lon)
                # -xsin(psi)+ycos(psi), psi = longitude (0 to 360)
                new_top = -xarr*sin((lon_arr+180.)*nppi/180.) +\
                    yarr*cos((lon_arr+180.)*nppi/180.)
            elif 'v' in varname:
                # meridional/north -> convert to theta_hat vector (latitude)
                # xcos(psi)cos(theta)+ysin(psi)cos(theta)
                # sin(theta) is always zero at the poles
                # theta = latitude (0 to 180), psi = longitude (0 to 360)
                new_top = xarr * cos((lon_arr+180.) * nppi/180.) *\
                    cos((90.-latval)*nppi/180.) + yarr *\
                    sin((lon_arr+180.) * nppi/180.) * cos((90.-latval) *
                                                          nppi/180.)

            # flip lon around so values match lon location (keep zero ref true)
            zero_idx = min(where(self._lon >= 0.)[0])  # find splitting index
            # create empty array for destination
            top = zeros((shape_list[0], shape_list[1]-1, shape_list[3]))
            div = (shape_list[1]-1) % 2  # deal with even or odd number of lon
            # move last half to first half and vice versa
            top[:, :zero_idx-div, :] = new_top[:, zero_idx:, :]
            top[:, zero_idx-div:, :] = new_top[:, :zero_idx, :]
            return top

        def register_variable(self, varname, gridded_int, verbose=False):
            '''register the variable data with the chunked interpolation method
            since the data files are small and chunked.'''

            # determine coordinate variables and xvec by coord list
            gvar = self.variables[varname]['data']
            p = list(self.pattern_files.keys())[0]
            coord_list = [value[5] for key, value in model_varnames.items()
                          if value[0] == varname][0]
            # time grids are the same in both hemispheres
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[p]['all']}}
            if 'lon' in coord_list:
                coord_dict['lon'] = {'units': 'deg', 'data': self._lon}
                coord_dict['lat'] = {'units': 'deg', 'data': self._lat}
            if 'milev' in coord_list and hasattr(self, '_milev'):
                coord_dict['mlon'] = {'units': 'deg', 'data': self._mlon}
                coord_dict['mlat'] = {'units': 'deg', 'data': self._mlat}
                coord_dict['milev'] = {'units': 'm/m', 'data': self._milev}
            if 'ilev1' in coord_list and hasattr(self, '_ilev1'):
                coord_dict['ilev1'] = {'units': 'm/m', 'data': self._ilev1}
            elif 'ilev' in coord_list and hasattr(self, '_ilev'):
                coord_dict['ilev'] = {'units': 'm/m', 'data': self._ilev}
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]

            # define data retrieval and wrangling logic
            # keep in mind that high-alt files have one time step per file
            # normal files have multiple time steps per file
            def func(i):  # i is the file number
                # get the data
                file = self.pattern_files[p][i]
                cdf_data = Dataset(file)
                data = array(cdf_data.variables[gvar])
                cdf_data.close()
                if data.shape[0] > 1 and file != self.pattern_files[p][-1]:
                    # if not the last file, tack on first time from next
                    next_file = self.pattern_files[p][i+1]
                    cdf_data = Dataset(next_file)
                    data_slice = array(cdf_data.variables[gvar][0])
                    cdf_data.close()
                    data = append(data, [data_slice], axis=0)

                # wrangle the data
                if len(data.shape) == 3:
                    # (t,lat,lon) -> (t,lon,lat)
                    variable = transpose(data, (0, 2, 1))
                    out = self.wrap_3Dlatlon(varname, variable)
                    if out.shape[0] == 1:  # high-alt data has one time
                        return squeeze(out)
                    else:
                        return out
                # 4D specific logic from here on down
                # (t,h,lat,lon) -> (t,lon,lat,h)
                variable = transpose(data, (0, 3, 2, 1))
                if 'lat' in coord_list:
                    out = self.wrap_4Dlatlon(varname, variable)
                    if out.shape[0] == 1:  # high-alt data has one time
                        return squeeze(out)
                    else:
                        return out
                # otherwise, look for undefined top rows and remove them
                # they need to be removed to avoid interpolation problems
                # only occurs in mlon/mlat/milev dependent variables (H_imlev)
                top_idx = -1
                top_shape = list(variable[:, :, :, top_idx].shape)
                top_size = top_shape[0] * top_shape[1] * top_shape[2]  # 3D arr
                idx_top = where(variable[:, :, :, top_idx] > 1e+35)[0]
                while top_size == len(idx_top):  # replace undefined top row(s)
                    if verbose:
                        print('All values at max milev are 1e+36 for ' +
                              f'{varname}. Slicing off top array.')
                    variable[:, :, :, top_idx] = NaN
                    top_idx -= 1
                    idx_top = where(variable[:, :, :, top_idx] > 1e+35)[0]
                if variable.shape[0] == 1:  # high-alt data has one time
                    return squeeze(variable)
                else:
                    return variable

            # determine interpolation method based on file structuring
            if len(self.times[p]['all']) > len(self.times[p]['start']):
                # time chunking method
                interp_flag = 2
            elif len(self.times[p]['all']) == len(self.times[p]['start']):
                # time slicing method
                interp_flag = 1

            # need H functions to be gridded regardless of gridded_int value
            h_grid = True if varname in ['H_ilev', 'H_ilev1'] else gridded_int
            self = RU.Functionalize_Dataset(
                self, coord_dict, varname, self.variables[varname],
                h_grid, coord_str, interp_flag=interp_flag, func=func,
                times_dict=self.times[p])

            # perform H_ilev/H_ilev1 substitution if needed
            if isinstance(self.ilev_sub, str) and varname == self.ilev_sub:
                other_name = ['H_ilev', 'H_ilev1']
                other_name.remove(varname)  # first element is the other name
                print(f'{other_name[0]} missing in data and is needed to ' +
                      'convert the requested variables to depend on height.' +
                      f' Using {varname} instead.')
                del coord_dict[other_name[0]]
                # e.g. set ilev1 grid equal to ilev grid if H_ilev1 is missing
                coord_dict[varname[2:]] = {
                    'units': 'm/m', 'data': getattr(self, other_name[0][2:])}
                coord_units = {key: value['units'] for key, value in
                               coord_dict.items()}
                self.variables[other_name[0]] = self.variables[varname]
                # register the other variable by linking to this one
                self = RU.register_interpolator(self, other_name[0], varname,
                                                coord_units)

            # create pressure level -> km function once per ilev type
            if varname in ['H_ilev', 'H_ilev1'] or varname in self.total_ilev:
                if varname in ['H_ilev', 'H_ilev1']:  # create custom interp
                    new_varname = 'P'+coord_list[-1][1:]
                    kms = getattr(self, '_km_'+coord_list[-1])
                    kms_max = getattr(self, '_km_'+coord_list[-1]+'_max')
                    kms_min = getattr(self, '_km_'+coord_list[-1]+'_min')
                    # perform unit conversion if needed
                    if self.variables[varname]['units'] != 'km':
                        self[varname+'km_ijk[km]'] = varname + '_ijk'
                        km_interp = self[varname+'km_ijk']
                    else:
                        km_interp = self[varname+'_ijk']
                    # Import and call custom interpolator
                    units = 'm/m'
                    self[new_varname], interp_ijk = RU.PLevelInterp(
                        km_interp, coord_dict['time']['data'],
                        coord_dict['lon']['data'], coord_dict['lat']['data'],
                        coord_dict[coord_list[-1]]['data'], units, kms,
                        [kms_min, kms_max])
                    # kms is a 1D array of the median height values in km
                else:  # define by function composition
                    new_varname = varname.split('_ilev')[0]
                    units = self.variables[varname]['units']
                    # substitute kms array if height was also substituted
                    if isinstance(self.ilev_sub, str):
                        other_name = ['ilev', 'ilev1']
                        other_name.remove(self.ilev_sub[2:])
                        kms = getattr(self, '_km_'+other_name[0])
                        kms_max = getattr(self, '_km_'+other_name[0]+'_max')
                        kms_min = getattr(self, '_km_'+other_name[0]+'_min')
                        self[new_varname] = varname+'(P'+other_name[0][1:]+')'
                        interp_ijk = self[new_varname]
                        self[new_varname].meta['arg_units'] = \
                            self['P'+other_name[0][1:]].meta['arg_units']
                    else:
                        kms = getattr(self, '_km_'+coord_list[-1])
                        kms_max = getattr(self, '_km_'+coord_list[-1]+'_max')
                        kms_min = getattr(self, '_km_'+coord_list[-1]+'_min')
                        self[new_varname] = varname+'(P'+coord_list[-1][1:]+')'
                        interp_ijk = self[new_varname]
                        self[new_varname].meta['arg_units'] = \
                            self['P'+coord_list[-1][1:]].meta['arg_units']
                self.variables[new_varname] = {'units': units}

                # create gridded interpolator if requested
                if h_grid:
                    self = RU.register_griddedPlev(self, new_varname, units,
                                                   interp_ijk, coord_dict, kms,
                                                   [kms_min, kms_max])
            return
    return MODEL



# -*- coding: utf-8 -*-
"""
Created on Thu May 13 18:28:18 2021
@author: rringuet
"""
from kamodo import kamodofy, gridify
from numpy import NaN, vectorize, append, array, meshgrid, ravel, diff
from numpy import unique, zeros, ndarray, floor, float32, all, insert
from scipy.interpolate import RegularGridInterpolator as rgiND
from scipy.interpolate import interp1d as rgi1D
import forge
from datetime import datetime, timezone, timedelta
from os.path import basename


def create_interp(coord_data, data_dict, func=None, func_default='data'):
    '''Create an interpolator depending on the dimensions of the input.
    Inputs:
        coord_data: a dictionary containing the coordinate data.
            {'name_of_coord1': coord1_data', 'name_of_coord2': 'coord2_data',
             etc...}. All arrays should be 1D arrays.
        data_dict: a dictionary containing the data information.
            {'units': 'data_units', 'data': data_array}
            data_array should have the same shape as (c1, c2, c3, ..., cN)
        Note: The dataset must also depend upon ALL of the coordinate arrays
            given.
    Output: an interpolator created with the given dataset and coordinates.
    '''

    # determine the number of coordinates, create the interpolator
    coord_list = [value for key, value in coord_data.items()]  # list of arrays
    n_coords = len(coord_data.keys())
    if n_coords == 1 and func_default == 'data':
        rgi = rgi1D(*coord_list, data_dict['data'], bounds_error=False,
                    fill_value=NaN)
        # wrap in a function and return the function
        def interp(xvec):
            return rgi(xvec)
        return interp
    elif n_coords > 1 and func_default == 'data':
        rgi = rgiND(coord_list, data_dict['data'], bounds_error=False,
                    fill_value=NaN)
        # wrap in a function and return the function
        def interp(xvec):
            return rgi(xvec)
        return interp
    elif func_default == 'custom':
        return func  # returns the custom interpolator


def create_funcsig(coord_data, coord_str, bounds):
    '''Create a custom function signature based on the dimensions and the
    coordinate string given (e.g. "SMcar").
    Inputs:
        coord_data: a dictionary containing the coordinate data.
            {'name_of_coord1': coord1_data', 'name_of_coord2': 'coord2_data',
             etc...}. All arrays should be 1D arrays.
        coord_str: a string indicating the coordinate system of the data
            (e.g. "SMcar" or "GEOsph").
        bounds: an array of positions for the max and min values of each
            coordinate grid transposed.
            aa, bb, cc, dd = np.meshgrid(trange, lon_range, lat_range, h_range)
            a, b, c, d = np.ravel(aa), np.ravel(bb), np.ravel(cc), np.ravel(dd)
            bounds = array([a, b, c, d]).T
            where trange, lon_range, ... = [t_min, t_max], [lon_min, lon_max]
    Outputs: A forge parameter object.
    '''

    # determine the number of coordinates
    n_coords = len(coord_data.keys())
    if n_coords == 1:
        # prepare the custom function signature
        if coord_str != '':
            coord_name = list(coord_data.keys())[0]+'_'+coord_str
        else:
            coord_name = list(coord_data.keys())[0]
    else:
        # prepare the custom function signature
        if 'sph' in coord_str:
            coord_name = 'rvec_'+coord_str+str(n_coords)+'D'
        else:
            coord_name = 'xvec_'+coord_str+str(n_coords)+'D'
    # e.g. 'xvec_SMcar4D' or 'rvecGEOsph3D'
    param_xvec = forge.FParameter(
        name=coord_name, interface_name='xvec',
        kind=forge.FParameter.POSITIONAL_OR_KEYWORD,
        default=bounds)
    return param_xvec


def Functionalize_Dataset(kamodo_object, coord_dict, variable_name,
                          data_dict, gridded_int, coord_str, interp_flag=0,
                          func=None, times_dict=None, func_default='data'):
    '''Determine and call the correct functionalize routine.
    Inputs:
        kamodo_object: the previously created kamodo object.
        coord_dict: a dictionary containing the coordinate information.
            {'name_of_coord1': {'units': 'coord1_units', 'data': coord1_data},
             'name_of_coord2': {'units': 'coord2_units', 'data': coord2_data},
             etc...}
            coordX_data should be a 1D array. All others should be strings.
        variable_name: a string giving the LaTeX representation of the variable
        data_dict: a dictionary containing the data information.
            {'units': 'data_units', 'data': data_array}
            If interp_flag=0 is used, data_array should be a numpy array of
            shape (c1, c2, c3, ..., cN). Otherwise, it could be a string or
            other thing that feeds into the given func.
        Note: The dataset must depend upon ALL of the coordinate arrays given.

        gridded_int: True to create a gridded interpolator (necessary for
            plotting and slicing). False otherwise (e.g. for the flythrough).
        coord_str: a string indicating the coordinate system of the data
            (e.g. "SMcar" or "GEOsph").
        interp_flag: the chosen method of interpolation.
            Options:
                0: (default option) assumes the given data_dict['data'] is an
                    N-dimensional numpy array and creates a standard scipy
                    interpolator to functionalize the data. This option should
                    be used when the entire dataset is contained in a single
                    array AND can easily fit into memory (e.g. 1D time series)
                1: Lazy interpolation is applied on top of the standard scipy
                    interpolators (one per time slice). This option is designed
                    to be used when the data is stored in one time step per
                    file and the time step can easily fit into memory.
                2: Lazy chunked interpolation is applied on top of the
                    standard scipy interpolators (one per time chunk).
                    Interpolation between time chunks occurs in the given
                    func. This option is designed to be used when the data is
                    stored with more than one time step per file AND the time
                    chunks are small enough to easily fit into memory.
                3. Combination of 1 and 2, used when the time chunks are too
                    large for the typical computer memory. Searches for the
                    correct time chunk, then the correct time slice in the
                    time chunk. This option is designed to be used when the
                    files contain more than one time step and are too large to
                    easily read into memory.
        func: a function defining the logic to be executed on a given time
            slice or chunk (e.g. converting to an array and then transposing
            it).
            *** Only needed for interp_flag greater than zero. ***
            - For interp_flag=1, the function should return only the data for
                the time slice. Required syntax is data = func(i), where i is
                the index of the slice on the full time grid.
            - For interp_flag=2, the function should return the data for the
                time chunk. Required syntax is data = func(i), where i is the
                file number in a list of files of the same naming pattern.
            - For interp_flag=3, the function should return only the data for
                the time slice. Required syntax is data = func(i, fi), where i
                is the file number in a list of files of the same naming
                pattern, and fi is the index of the time slice in that file.
            - ALL functions must include the method to retrieve the data from
                the file.
            - This option is useful when a dataset's shape order does not match
            the required order (e.g. t, lat, lon -> t, lon, lat) and allows
            such operations to be done on the fly.
        start_times: a list of the start times for each data chunk.
            *** Only needed for interpolation over time chunks. ***
            (interp_flag options 2 and 3).
        func_default: a string indicating the type of object returned by func.
            The default is 'data', indicating that the returned object is a
            numpy array. Set this to 'custom' to indicate that func returns
            a custom interpolator.

    Output: A kamodo object with the functionalized dataset added.
    '''

    # split the coord_dict into data and units
    coord_data = {key: value['data'] for key, value in coord_dict.items()}
    coord_units = {key: value['units'] for key, value in coord_dict.items()}

    # create the bounds array (see docstring of create_funcsig)
    # positions of corners of coordinate box
    data_list = [array([value['data'].min(), value['data'].max()]) for
                 key, value in coord_dict.items()]
    mesh_list = meshgrid(*data_list)
    bounds = array([ravel(item) for item in mesh_list], dtype=float).T

    # create a functionalized interpolator and modified function signature
    param_xvec = create_funcsig(coord_data, coord_str, bounds)
    if interp_flag == 0:  # standard logic
        interp = create_interp(coord_data, data_dict, func,
                               func_default=func_default)
    elif interp_flag == 1:
        interp = time_interp(coord_dict, data_dict, func,
                             func_default=func_default)
    elif interp_flag == 2:
        interp = multitime_interp(coord_dict, data_dict, times_dict, func,
                                  func_default=func_default)
    elif interp_flag == 3:
        interp = multitime_biginterp(coord_dict, data_dict, times_dict, func,
                                     func_default=func_default)
    new_interp = forge.replace('xvec', param_xvec)(interp)
    interp = kamodofy(units=data_dict['units'], data=data_dict['data'],
                      arg_units=coord_units)(new_interp)

    # add gridded version if requested, even for 1D functions
    kamodo_object[variable_name] = interp
    if gridded_int:
        interp_grid = kamodofy(gridify(interp, **coord_data),
                               units=data_dict['units'],
                               data=data_dict['data'], arg_units=coord_units)
        kamodo_object[variable_name+'_ijk'] = interp_grid
    return kamodo_object


def time_interp(coord_dict, data_dict, func, func_default='data'):
    '''Create a functionalized interpolator by splitting into timesteps.
    interp_flag=1 uses this function.
    Inputs:
        coord_dict: a dictionary containing the coordinate information.
            {'name_of_coord1': {'units': 'coord1_units', 'data': coord1_data},
             'name_of_coord2': {'units': 'coord2_units', 'data': coord2_data},
             etc...}
            coordX_data should be a 1D array. All others should be strings.
        data_dict: a dictionary containing the data information.
            {'units': 'data_units', 'data': data_array}
            data_array should have the same shape as (c1, c2, c3, ..., cN)
            Note: The dataset must also depend upon ALL of the coordinate
                arrays given.
        func: a function defining the logic to be executed on a given time
            slice i, which must include data retrieval from the file. The
            should return the data for the time slice. Required syntax is
            data = func(i).
        func_default: a string indicating the type of object returned by func.
            The default is 'data', indicating that the returned object is a
            numpy array. Set this to 'custom' to indicate that func returns
            a custom interpolator.
    Output: A lazy time interpolator.
    '''

    # create a list of interpolators per timestep, not storing the data, and
    # interpolate from there. such as done in superdarnea_interp.py
    # Assumes that time is the first dimension

    # split the coord_dict into data and units
    coord_list = [value['data'] for key, value in coord_dict.items()]  # arrays
    idx_map, time_interps = [], []  # initialize variables

    # define method for how to add time slices to memory
    def add_timeinterp(i, time_interps):
        if len(coord_list) > 1 and func_default == 'data':
            time_interps.append(rgiND(coord_list[1:], func(i),
                                      bounds_error=False, fill_value=NaN))
        elif len(coord_list) == 1:  # has to be different for time series data
            def dummy_1D(spatial_position):
                return func(i)  # data only
            time_interps.append(dummy_1D[0])
        elif func_default == 'custom':
            time_interps.append(func(i))
        return time_interps

    def interp_i(*args, time_interps=time_interps, idx_map=idx_map):

        position = array([*args])  # time coordinate value must be first
        times = position[0]
        sposition = position[1:].T  # spatial coordinates (len(times), ncoords)
        if len(position.shape) == 1:
            out_vals = zeros(1) * NaN
        else:
            out_vals = zeros(times.shape[0]) * NaN
        # loop through time grid instead of input time array
        for i in range(len(coord_list[0])):
            # figure out what times given are in this file, if any
            if i < len(coord_list[0]) - 1:
                end_time, idx_list = coord_list[0][i+1], [i, i+1]
            else:
                end_time, idx_list = coord_list[0][i], [i]
            if isinstance(times, ndarray):
                st_idx = [j for j, time_val in enumerate(times) if
                          time_val >= coord_list[0][i] and
                          time_val <= end_time]
            else:
                if times >= coord_list[0][i] and times <= end_time:
                    st_idx = [i]
                else:
                    continue
            if len(st_idx) == 0:  # skip if none
                continue

            # Interpolate values for chosen time grid values
            for i in idx_list:
                if i not in idx_map:
                    len_map = len(idx_map)  # length before adding another
                    try:  # allow for memory errors
                        time_interps = add_timeinterp(i, time_interps)
                    except MemoryError:  # remove one time slice first
                        print('Avoiding memory error...')
                        if len_map == 0:
                            print('Not enough memory to load one time slice.' +
                                  ' Please close some applications.')
                        elif len_map < 2:  # failed to add a second slice
                            print('Not enough memory to load two time slices' +
                                  '. Please close some applications.')
                        elif abs(i-idx_map[0]) > 2:
                            del idx_map[0], time_interps[0]
                        else:
                            del idx_map[-1], time_interps[-1]
                        time_interps = add_timeinterp(i, time_interps)
                    print(f'Time slice index {i} added from file.')
                    idx_map.append(i)
            if len(idx_list) > 1:
                interp_locations = [idx_map.index(val) for val in idx_list]
                if isinstance(times, ndarray):
                    interp_values = array([time_interps[i](sposition[st_idx])
                                           for i in interp_locations]).T
                    for j, vals in enumerate(interp_values):  # loop positions
                        time_int = rgi1D(coord_list[0][idx_list], vals,
                                         bounds_error=False,
                                         fill_value=NaN)
                        out_vals[st_idx[j]] = time_int(times[st_idx[j]])
                else:
                    interp_values = array([time_interps[i](sposition)
                                           for i in interp_locations]).T
                    time_int = rgi1D(coord_list[0][idx_list],
                                     interp_values,
                                     bounds_error=False, fill_value=NaN)
                    out_vals = time_int(times)
            else:
                interp_location = idx_map.index(idx_list[0])
                if isinstance(times, ndarray):
                    out_vals[st_idx] = time_interps[interp_location](
                        sposition[st_idx])
                else:
                    out_vals = time_interps[interp_location](sposition)
        return out_vals

    def interp(xvec):
        return interp_i(*array(xvec).T)
    return interp


def multitime_interp(coord_dict, data_dict, times_dict, func,
                     func_default='data'):
    '''Create a functionalized interpolator by splitting into time chunks.
    Inputs:
        coord_dict: a dictionary containing the coordinate information.
            {'name_of_coord1': {'units': 'coord1_units', 'data': coord1_data},
             'name_of_coord2': {'units': 'coord2_units', 'data': coord2_data},
             etc...}  # time not given
            coordX_data should be a 1D array. All others should be strings.
        data_dict: a dictionary containing the data information.
            {'units': 'data_units', 'data': empty}
            data_array should have the same shape as (c1, c2, c3, ..., cN)
            Note: The dataset must also depend upon ALL of the coordinate
                arrays given.
        start_times: the start times of each file in the file_dir.
            Interpolation between time chunks should be taken care of by
            appending the first time slice of the next time chunk to the end of
            the current time chunk (see func description).
        func: a function defining the logic to be executed on a given time
            chunk, including retrieving the data from the file. The function
            must return the data for the time chunk. Interpolation between time
            chunks can be easily achieved by appending a time slice from the
            next chunk to the current one in the logic of the function. The
            required syntax is data = func(i), where i is the file number.
        func_default: a string indicating the type of object returned by func.
            The default is 'data', indicating that the returned object is a
            numpy array. Set this to 'custom' to indicate that func returns
            a custom interpolator.
    Output: A time-chunked lazy interpolator.
    '''

    # create a list of interpolators per time chunk, not storing the data, and
    # interpolate from there. such as done in superdarnea_interp.py
    # Assumes that time is the first dimension.
    # Assumes that data_dict['data'][i] is a string used to find the right file

    # split the coord_dict into data and units, initialize variables/func
    coord_list = [value['data'] for key, value in coord_dict.items()]  # arrays
    idx_map, time_interps = [], []  # initialize variables

    # define method for how to add time chunks to memory
    def add_timeinterp(i, time_interps):  # i is the file number
        idx_map.append(i)
        # determine time grid for file
        start_idx = list(coord_list[0]).index(times_dict['start'][i])
        end_idx = list(coord_list[0]).index(times_dict['end'][i]) + 1
        if i < len(times_dict['end'])-1:  # interpolating btwn files
            end_idx += 1
        time = coord_list[0][start_idx:end_idx]
        # get data for chunk
        data = func(i)
        if len(coord_list) > 1 and func_default == 'data':
            coord_list_i = [time] + coord_list[1:]
            time_interps.append(rgiND(coord_list_i, data,
                                      bounds_error=False, fill_value=NaN))
        elif len(coord_list) == 1:  # has to be different for time series data
            time_interps.append(rgi1D(time, data, bounds_error=False,
                                      fill_value=NaN))
        elif func_default == 'custom':  # when func returns an interpolator
            time_interps.append(func(i))
        return time_interps, idx_map

    def interp_i(*args, time_interps=time_interps, idx_map=idx_map):
        position = array([*args])  # time coordinate value must be first
        if len(position.shape) == 1:
            out_vals = zeros(1) * NaN
        else:
            out_vals = zeros(position.shape[1]) * NaN  # (num_cgrids, num_pos)
        # loop through start times instead
        for i in range(len(times_dict['start'])):
            # figure out what times given are in this file, if any
            if i < len(times_dict['start']) - 1:
                end_time = times_dict['start'][i+1]
            else:
                end_time = times_dict['end'][i]
            if isinstance(position[0], ndarray):
                st_idx = [j for j, time_val in enumerate(position[0]) if
                          time_val >= times_dict['start'][i] and
                          time_val <= end_time]
            else:
                if position[0] >= times_dict['start'][i] and \
                        position[0] <= end_time:
                    st_idx = [i]
                else:
                    continue
            if len(st_idx) == 0:  # skip if none
                continue

            # Interpolate values for chosen time chunk
            if i not in idx_map:
                len_map = len(idx_map)
                try:  # allow for memory errors
                    time_interps, idx_map = add_timeinterp(i, time_interps)
                except MemoryError:  # remove two(?) items first
                    print('Avoiding memory error...')
                    if len_map == 0:  # tried but failed
                        print('Not enough memory to load a time chunk. ' +
                              'Please close some applications.')
                    elif i != idx_map[0]:
                        del idx_map[0], time_interps[0]
                    else:
                        del idx_map[-1], time_interps[-1]
                    time_interps, idx_map = add_timeinterp(i, time_interps)
                    print(f'Time chunk added from file {i+1}.')
            # interpolate all the positions with times in this time chunk
            if isinstance(position[0], ndarray):
                out_vals[st_idx] = time_interps[idx_map.index(i)](
                    position[:, st_idx].T)
            else:
                out_vals = time_interps[idx_map.index(i)](position)
        return out_vals

    def interp(xvec):
        return interp_i(*array(xvec).T)

    return interp


def multitime_biginterp(coord_dict, data_dict, times_dict, func,
                        func_default='data'):
    '''Create a functionalized interpolator by splitting into time chunks as
    stored in the files and then into time steps. Use this option (3) when the
    time chunks are large compared to the typically available computer memory.
    Inputs:
        coord_dict: a dictionary containing the coordinate information.
            {'name_of_coord1': {'units': 'coord1_units', 'data': coord1_data},
             'name_of_coord2': {'units': 'coord2_units', 'data': coord2_data},
             etc...}  # time not given
            coordX_data should be a 1D array. All others should be strings.
        data_dict: a dictionary containing the data information.
            {'units': 'data_units', 'data': empty}
            data_array should have the same shape as (c1, c2, c3, ..., cN)
            Note: The dataset must also depend upon ALL of the coordinate
                arrays given.
        start_times: the start times of each file in the file_dir.
            Interpolation between time chunks should be taken care of by
            appending the first time slice of the next time chunk to the end of
            the current time chunk (see func description).
        func: a function defining the logic to be executed on a given time
            chunk, including retrieving the correct time slice of data from the
            file. The function must return the data for the time SLICE.
            Required syntax is data = func(f, i), where f is the file number
            of the time chunk and i is the time slice number in that file.
            Counting starts at zero.
        func_default: a string indicating the type of object returned by func.
            The default is 'data', indicating that the returned object is a
            numpy array. Set this to 'custom' to indicate that func returns
            a custom interpolator.
    Output: A time-chunked lazy interpolator that loads two slices each time.
    '''

    # create a list of interpolators per time chunk, not storing the data, and
    # interpolate from there. such as done in superdarnea_interp.py
    # Assumes that time is the first dimension.
    # Assumes that data_dict['data'][i] is a string used to find the right file

    # split the coord_dict into data and units
    coord_list = [value['data'] for key, value in coord_dict.items()]  # arrays
    idx_map, time_interps = [], []  # initialize variables

    # define method for how to add time slices to memory
    # i is the file#, fi is the time slice # in file
    def add_timeinterp(i, fi, time_interps):
        if len(coord_list) > 1 and func_default == 'data':
            time_interps.append(rgiND(coord_list[1:], func(i, fi),
                                      bounds_error=False, fill_value=NaN))
        # has to be different for time series data
        elif len(coord_list) == 1 and func_default == 'data':
            def dummy_1D(spatial_position):
                return func(i, fi)  # data only
            time_interps.append(dummy_1D[0])
        elif func_default == 'custom':
            time_interps.append(func(i, fi))
        return time_interps

    def interp_i(*args, time_interps=time_interps, idx_map=idx_map):

        position = array([*args])  # time coordinate value must be first
        times = position[0]
        sposition = position[1:].T  # spatial coordinates (len(times), ncoords)
        if len(position.shape) == 1:
            out_vals = zeros(1) * NaN
        else:
            out_vals = zeros(times.shape[0]) * NaN
        # loop through time grid instead of input time array
        for ti in range(len(coord_list[0])):
            # figure out what times given are in this slice, if any
            if ti < len(coord_list[0]) - 1:
                end_time, idx_list = coord_list[0][ti+1], [ti, ti+1]
            else:
                end_time, idx_list = coord_list[0][ti], [ti]
            if isinstance(times, ndarray):
                st_idx = [j for j, time_val in enumerate(times) if
                          time_val >= coord_list[0][ti] and
                          time_val <= end_time]
            else:
                if times >= coord_list[0][ti] and times <= end_time:
                    st_idx = [ti]
                else:
                    continue
            if len(st_idx) == 0:  # skip if none
                continue

            # Interpolate values for chosen time grid values
            for idx in idx_list:
                if idx not in idx_map:
                    # Get index of file containing the desired time
                    i = get_file_index(coord_list[0][idx], times_dict['start'])
                    # determine which slice in the file corresponds to time val
                    # aka time position relative to beginning of file
                    start_idx = list(coord_list[0]).index(
                        times_dict['start'][i])
                    fi = list(coord_list[0][start_idx:]).index(
                        coord_list[0][idx])

                    # add time slice
                    len_map = len(idx_map)  # length before adding another
                    try:  # allow for memory errors
                        time_interps = add_timeinterp(i, fi, time_interps)
                    except MemoryError:  # remove one time slice first
                        print('Avoiding memory error...')
                        if len_map == 0:
                            print('Not enough memory to load one time slice.' +
                                  ' Please close some applications.')
                        elif len_map < 2:  # failed to add a second slice
                            print('Not enough memory to load two time slices' +
                                  '. Please close some applications.')
                        elif abs(idx-idx_map[0]) > 2:
                            del idx_map[0], time_interps[0]
                            del idx_map[0], time_interps[0]
                        else:
                            del idx_map[-1], time_interps[-1]
                            del idx_map[-1], time_interps[-1]
                        time_interps = add_timeinterp(i, fi, time_interps)
                    idx_map.append(idx)
                    print(f'Time slice index {idx} (file time {fi}) added ' +
                          f'from file {i+1}.')
            if len(idx_list) > 1:
                interp_locations = [idx_map.index(val) for val in idx_list]
                if isinstance(times, ndarray):
                    interp_values = array([time_interps[ii](sposition[st_idx])
                                           for ii in interp_locations]).T
                    for j, vals in enumerate(interp_values):  # loop positions
                        time_int = rgi1D(coord_list[0][idx_list], vals,
                                         bounds_error=False,
                                         fill_value=NaN)
                        out_vals[st_idx[j]] = time_int(times[st_idx[j]])
                else:  # one position
                    interp_values = array([time_interps[ii](sposition)
                                           for ii in interp_locations]).T
                    time_int = rgi1D(coord_list[0][idx_list],
                                     interp_values, bounds_error=False,
                                     fill_value=NaN)
                    out_vals = time_int(times)
            else:
                interp_location = idx_map.index(idx_list[0])
                if isinstance(times, ndarray):
                    out_vals[st_idx] = time_interps[interp_location](
                        sposition[st_idx])
                else:
                    out_vals = time_interps[interp_location](sposition)
        return out_vals

    def interp(xvec):
        return interp_i(*array(xvec).T)
    return interp


def get_slice_idx(time_val, time_array):
    '''Figures out where the time_val is in time_array.
    Inputs:
        time_val: float value of time
        time_array: array of float time values
    Returns: A list of indices for one time value on either side OR a list
        containing the single index that corresponds exactly to the given
        time_val.
    '''
    if time_val not in time_array:
        idx = sorted(append(time_array, time_val)).index(time_val)
        if idx > 1 and idx < len(time_array)-1:  # middle somewhere
            idx_list = [idx-1, idx]
        elif idx <= 1:  # beginning
            idx_list = [0, 1]
        elif idx > 1 and idx >= len(time_array)-1:  # at end
            idx_list = [len(time_array)-2, len(time_array)-1]
    else:
        idx = list(time_array).index(time_val)
        idx_list = [idx]
    return idx_list


def get_file_index(time_val, time_grid):
    '''Figures out which file start time is directly before the given time
    value.
    Inputs:
        time_val: float value of desired coordinate
        time_grid: an array of coordinate grid values of the same coordinate
    Returns: index of the start time directly preceding the given time value.
    '''

    if time_val not in time_grid:
        idx = sorted(append(time_grid, time_val)
                     ).index(time_val)  # get file #
        if idx == 0:
            i = 0  # beg of first file
        else:
            i = idx - 1  # somewhere else
    else:
        i = list(time_grid).index(time_val)
    return i


# ############# BEGIN PRESSURE LEVEL INVERSION ROUTINE #######################

# -*- coding: utf-8 -*-
# The pressure level inversion routine used by many ITM models


def PLevelInterp(h_func, time, longitude, latitude, ilev, units, km_grid,
                 km_min_max):
    '''Custom interpolator functionto convert from altitude in km to pressure
    level.
    Parameters:
        h_func - the H_ilev gridded function/interpolator in km
        time - a 1D array with the grid values for time
        longitude - a 1D array with the grid values for longitude
        latitude - a 1D array with the grid values for latitude
        ilev - a 1D array with the grid values for pressure level
        units - a string indicating the units of the pressure level grid
        km_grid - a 1D array of the median altitude values for each pressure
            level
        km_min_max - a 2-element list of the absolute min and max of the
            altitude in km

    Output: Two interpolators. The first interpolator accepts time, lon, lat,
        and pressure level and returns time, lon, lat, and height. The second
        interpolator accepts the same arguments and returns only the height.
        Both interpolators are 'kamodofied'.

    This code changes a function from being dependent on pressure level to
    instead depending upon altitude. The model reader calling this function
    needs to first create a gridded height variable using typical methods and
    use Kamodo-core's unit conversion capability to convert the gridded height
    function into km. This routine creates a function that uses the slicing
    technique to find the height values at a given time, longitude and latitude
    for the default grid values of pressure level. These height values and
    pressure level grid values are then used as the X and Y input values to a
    1D interpolator (height is X, pressure level is Y) to invert the function.
    This interpolator is then evaluated at the requested height (or heights)
    associated with the same time, longitude and latitude and returns the
    pressure level(s). The code repeats this process for each unique
    (t, lon, lat) trio. We use the interp1d interpolator for the 1D inversion
    (from SciPy). Regular and gridded interpolators based on this logic are
    returned.
    '''

    def km_to_ilev(t, lon, lat, km):
        '''Inputs t, lon, lat, and km are arrays;
        the interpolated pressure levels are returned.'''
        # sort lats and lons per time value
        input_arr = array([t, lon, lat]).T  # array of all positions given
        if not isinstance(t, ndarray):  # only one time/position
            km_vals = h_func(**{'time': input_arr[0], 'lon': input_arr[1],
                                'lat': input_arr[2]})
            km_interp = rgi1D(km_vals, ilev, bounds_error=False,
                              fill_value=NaN)
            return km_interp(km)

        # remaining logic is for if there is more than one position given
        pos_arr = unique(input_arr, axis=0)  # only create interp for unique
        out_ilev = zeros(len(t)) * NaN  # (t, lon, lat) positions to save time
        for pos in pos_arr:
            pos_idx = [i for i, p in enumerate(input_arr) if all(p == pos)]
            km_vals = h_func(**{'time': pos[0], 'lon': pos[1], 'lat': pos[2]})
            km_interp = rgi1D(km_vals, ilev, bounds_error=False,
                              fill_value=NaN)
            out_ilev[pos_idx] = km_interp(km[pos_idx])  # interp for all km
        return out_ilev

    # Convert f(ilev) to f(km)
    def plevconvert(xvec):
        '''Interpolator to convert from height to pressure level.
        Input xvec is a tuple or list of arrays/values.
        Returns the 4D position as a tuple of four 1D arrarys to feed into
        the original function.'''
        print('Inverting the pressure level grid. Please wait...', end="")
        t, lon, lat, km = array(xvec).T
        out_ilev = km_to_ilev(t, lon, lat, km)
        data = array([t, lon, lat, out_ilev]).T
        print('done.')
        return data

    def plevconvert_ijk(xvec):
        '''Interpolator to convert from height to pressure level.
        Input xvec is a tuple or list of arrays/values.
        Returns the 4D position as a tuple of four 1D arrarys to feed into
        the original function.'''
        print('Inverting the pressure level grid. Please wait...', end="")
        t, lon, lat, km = array(xvec).T
        data = km_to_ilev(t, lon, lat, km)
        print('done.')
        return data

    # split the coord_dict into data and units
    km_grid_altered = km_grid.copy()
    if diff(km_grid_altered).min() < 1:  # pressure grid inverted
        km_grid_altered[-1], km_grid_altered[0] = km_min_max
    else:
        km_grid_altered[0], km_grid_altered[-1] = km_min_max
    coord_data = {'time': time, 'lon': longitude, 'lat': latitude,
                  'height': km_grid_altered}
    coord_units = {'time': 'hr', 'lon': 'deg', 'lat': 'deg', 'height': 'km'}

    # create the bounds array (see docstring of create_funcsig)
    # positions of corners of coordinate box
    data_list = [array([value.min(), value.max()]) for key, value in
                 coord_data.items()]
    mesh_list = meshgrid(*data_list)
    bounds = array([ravel(item) for item in mesh_list], dtype=float).T

    # create the functionalized interpolators and modified function signatures
    fake_data = zeros((2, 2, 2, 2)) * NaN  # avoiding computation
    param_xvec = create_funcsig(coord_data, 'GDZsphkm', bounds)
    new_interp = forge.replace('xvec', param_xvec)(plevconvert)
    interp = kamodofy(units=units, data=fake_data, arg_units=coord_units
                      )(new_interp)
    new_interp = forge.replace('xvec', param_xvec)(plevconvert_ijk)
    interp_ijk = kamodofy(units=units, data=fake_data, arg_units=coord_units
                          )(new_interp)

    return interp, interp_ijk


def register_griddedPlev(kamodo_object, new_varname, units, interp_ijk,
                         coord_dict, kms, km_min_max):
    '''Properly register the gridded interpolators defined either from the
    pressure level inversion or by function composition. Need height coord
    grid to be median values in gridded creation but have absolute max/min in
    non-gridded creation.'''

    new_coord_units = {'time': 'hr', 'lon': 'deg',
                       'lat': 'deg', 'height': 'km'}
    fake_data = zeros((2, 2, 2, 2)) * NaN  # avoiding computation
    coord_data = {key: value['data'] for key, value in
                  coord_dict.items() if key in
                  new_coord_units.keys()}  # exclude ilev
    coord_data['height'] = insert(insert(kms, len(kms), km_min_max[-1]),
                                  0, km_min_max[0])  # add min and max
    kamodo_object.variables[new_varname+'_ijk'] = {'data': fake_data,
                                                   'units': units}
    kamodo_object[new_varname+'_ijk'] = kamodofy(
        gridify(interp_ijk, **coord_data),
        units=kamodo_object.variables[new_varname+'_ijk']['units'],
        data=kamodo_object.variables[new_varname+'_ijk']['data'],
        arg_units=new_coord_units)
    return kamodo_object


# ################ BEGIN TIME LIST FUNCTIONS #########################


@vectorize
def hrs_to_str(hrs, filedate):
    '''Convert hrs since midnight of first day to a string for the file list
    of format "Date: YYYY-MM-DD  Time: HH:MM:SS".'''
    h, m = floor(hrs), round(hrs % 1 * 60., 4)
    sec = round(((hrs - h) * 60. - m) * 60.)
    return datetime.strftime(filedate + timedelta(hours=int(h), minutes=int(m),
                                                  seconds=int(sec)),
                             '  Date: %Y-%m-%d  Time: %H:%M:%S')


@vectorize
def str_to_hrs(dt_str, filedate, format_string='%Y-%m-%d %H:%M:%S'):
    '''Convert datetime string of format "YYYY-MM-DD HH:MM:SS" to hrs since
    midnight of filedate.'''
    tmp = datetime.strptime(dt_str, format_string).replace(
        tzinfo=timezone.utc)
    return float32((tmp - filedate).total_seconds()/3600.)


@vectorize
def hrs_to_tstr(hrs, ms_timing=False):
    '''Convert number of hours into HH:MM:SS format. If ms_timing is True,
    str is in format HH:MM:SS.mms'''
    h, m = floor(hrs), round(hrs % 1 * 60., 4)
    if not ms_timing:
        sec = round(((hrs - h) * 60. - m) * 60.)
        return str(f'\n{int(h):02d}:{int(m):02d}:{int(sec):02d}')
    else:
        sec = ((hrs - h) * 60. - int(m)) * 60.
        sec_str = f'{sec:2.3f}'.zfill(6)
        return str(f'\n{int(h):02d}:{int(m):02d}:'+sec_str)


@vectorize
def tstr_to_hrs(time_str, ms_timing=False):
    '''Convert str from HH:MM:SS format to float 32. If ms_timing is True,
    str is in format HH:MM:SS.mms'''
    hh, mm, ss = time_str.split(':')
    t = float32(hh) + float32(mm)/60. + float32(ss)/3600.
    return float32(t)


def create_timelist(list_file, time_file, modelname, times, pattern_files,
                    filedate, ms_timing=False):
    '''Used by all the readers to create the time_file and list_file files.
    Inputs:
        list_file - the name of the file, including the file directory, to
            write the list of all of the files, start dates, start times,
            end dates, and end times.
        time_file - the name of the file, including the file directory, to
            write the time grid for each file pattern.
        modelname - a string with the name of the model
        times - a dictionary with keys indicating the file patterns.
            ['all'] - the complete time grid across the entire file_dir for
                the indicated file pattern.
            ['start'] - the starting time for each of the files of the given
                file pattern.
            ['end'] - the end time for each of the files of the given file
                pattern.
            All times are stored as the number of hours since midnight of the
                first file of all the files in the given directory.
        pattern_files - a dictionary with keys indicating the file patterns.
            Each key has as its value the list of files in the current file
            directory that match the file pattern.
        filedate - a datetime object indicating the date of the first file of
            all the files in the file_dir.
        ms_timing - a boolean indicating whether millisecond timing is needed.
            Default is False.
    Returns nothing.
    '''

    # create time list file if DNE
    print('Creating the time files...', end="")
    list_out = open(list_file, 'w')
    list_out.write(f'{modelname} file list start and end ' +
                   'dates and times')
    time_out = open(time_file, 'w')
    time_out.write(f'{modelname} time grid per pattern')
    for p in pattern_files.keys():
        # print out time grid to time file
        time_out.write('\nPattern: '+p)
        str_out = hrs_to_tstr(times[p]['all'], ms_timing)
        time_out.write(''.join(str_out))
        # print start and end dates and times to list file
        start_time_str = hrs_to_str(times[p]['start'], filedate)
        end_time_str = hrs_to_str(times[p]['end'], filedate)
        files = pattern_files[p]
        for i in range(len(files)):
            list_out.write('\n' + files[i].replace('\\', '/') +
                           start_time_str[i] + '  ' + end_time_str[i])
    time_out.close()
    list_out.close()
    print('done.')
    return


def read_timelist(time_file, list_file, ms_timing=False):
    '''Used by all readers to read in the time grid from the time_file and
    the list of files with start and end times from the list_file.
    Returns:
        times - a dictionary with keys indicating the file patterns.
            ['all'] - the complete time grid across the entire file_dir for
                the indicated file pattern.
            ['start'] - the starting time for each of the files of the given
                file pattern.
            ['end'] - the end time for each of the files of the given file
                pattern.
            All times are stored as the number of hours since midnight of the
                first file of all the files in the given directory.
        pattern_files - a dictionary with keys indicating the file patterns.
            Each key has as its value the list of files in the current file
            directory that match the file pattern.
        filedate - a datetime object indicating the date of the first file of
            all the files in the file_dir.
        filename - a string containing the names of all the files in the file
            directory, separated by commas.
        ms_timing - a boolean indicating whether millisecond timing is needed.
            Default is False.
    '''

    # get time grids and initialize self.times structure
    times, pattern_files = {}, {}
    time_obj = open(time_file)
    data = time_obj.readlines()
    for line in data[1:]:
        if 'Pattern' in line:
            p = line.strip()[9:]
            times[p] = {'start': [], 'end': [], 'all': []}
        else:
            times[p]['all'].append(line.strip())
    time_obj.close()
    for p in times.keys():
        times[p]['all'] = tstr_to_hrs(times[p]['all'], ms_timing)

    # get filenames, dates and times from list file
    files, start_date_times, end_date_times = [], [], []
    list_obj = open(list_file)
    data = list_obj.readlines()
    for line in data[1:]:
        file, tmp, date_start, tmp, start_time, tmp, date_end, \
            tmp, end_time = line.strip().split()
        files.append(file)
        start_date_times.append(date_start+' '+start_time)
        end_date_times.append(date_end+' '+end_time)
    list_obj.close()
    filedate = datetime.strptime(start_date_times[0][:10], '%Y-%m-%d').replace(
        tzinfo=timezone.utc)
    for p in times.keys():
        pattern_files[p] = [f for f in files if p in basename(f)]
        start_times = [t for f, t in zip(files, start_date_times)
                       if p in f]
        times[p]['start'] = str_to_hrs(start_times, filedate)
        end_times = [t for f, t in zip(files, end_date_times)
                     if p in f]
        times[p]['end'] = str_to_hrs(end_times, filedate)
    filename = ''.join([f+',' for f in files])[:-1]

    return times, pattern_files, filedate, filename



'''
Written by Rebecca Ringuette, 2022
'''

# variable name in file: [standardized variable name, descriptive term, units]
model_varnames = {'PHI': ['Phi', 'Electric potential', 0, 'SM', 'sph',
                          ['time', 'lon', 'lat'], 'kV'],
                  'Psi': ['Psi', 'Magnetic potential', 0, 'SM', 'sph',
                          ['time', 'lon', 'lat'], 'cT*m'],
                  'FAC': ['j_FAC', 'Field-aligned current density', 0, 'SM',
                          'sph', ['time', 'lon', 'lat'], 'muA/m**2'],
                  'JH': ['JH', 'Joule heating rate', 0, 'SM', 'sph',
                         ['time', 'lon', 'lat'], 'mW/m**2']}


def MODEL():

    from kamodo import Kamodo
    from netCDF4 import Dataset
    from glob import glob
    from os.path import basename, isfile
    from numpy import array
    from time import perf_counter
    from datetime import datetime, timezone
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''Weimer model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.
        Note: If you want to add a file to a directory already converted to nc,
            you need to delete the converted file and the _list.txt and
            _times.txt files to trigger a new file conversion on the next
            execution.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - Weimer model outputs are given in ascii files with one timestep
              per file. Since the files are extremely small, all of the data
              for the run is collected into one netCDF4 file.
            - The converted file is assumed to be larger than 16 GB, so
              interpolation method 3 is chosen. The standard SciPy
              interpolator is used.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     printfiles=False, filetime=False, gridded_int=True,
                     verbose=False, **kwargs):
            super(MODEL, self).__init__(**kwargs)
            self.modelname = 'Weimer'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # check for nc files
                nc_files = sorted(glob(file_dir+'*.nc'))
                txt_files = sorted(glob(file_dir+'*.txt'))
                if len(nc_files) == 0:  # perform file conversion if none
                    from kamodo_ccmc.readers.weimer_tocdf import convert_all
                    convert_all(txt_files)
                    nc_files = sorted(glob(file_dir+'*.nc'))
                self.filename = ''.join([f+',' for f in txt_files])[:-1]
                p = basename(nc_files[0]).split('.')[0]  # only one nc file

                # datetime object for midnight on date from first text file
                date = basename(txt_files[0])[-17:-9]  # YYYYMMDD
                self.filedate = datetime.strptime(
                    date+' 00:00:00', '%Y%m%d %H:%M:%S').replace(
                        tzinfo=timezone.utc)

                # establish time attributes
                # store list of files to loop through later
                self.pattern_files[p] = nc_files
                self.times[p] = {'start': [], 'end': [], 'all': []}

                # get times from single nc file
                cdf_data = Dataset(nc_files[0])
                tmp = array(cdf_data.variables['time'])  # hrs since midnit
                self.times[p]['start'] = array([tmp[0]])
                self.times[p]['end'] = array([tmp[-1]])
                self.times[p]['all'] = tmp
                cdf_data.close()

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            # return time info
            if filetime:
                return

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item
                            not in test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # collect variable list (in attributes of datasets)
            p = list(self.pattern_files.keys())[0]  # only one pattern
            cdf_data = Dataset(self.pattern_files[p][0])

            # get coordinates from first file
            self._lat = array(cdf_data.variables['lat'])
            self._lon = array(cdf_data.variables['lon'])

            # check var_list for variables not possible in this file set
            if len(variables_requested) > 0 and\
                    variables_requested != 'all':
                gvar_list = [key for key in model_varnames.keys()
                             if key in cdf_data.variables.keys() and
                             model_varnames[key][0] in variables_requested]
                if len(gvar_list) != len(variables_requested):
                    err_list = [value[0] for key, value in
                                model_varnames.items()
                                if key not in cdf_data.variables.keys() and
                                value[0] in variables_requested]
                    if len(err_list) > 0:
                        print('Some requested variables are not available: ',
                              err_list)
            else:
                gvar_list = [key for key in model_varnames.keys()
                             if key in cdf_data.variables.keys()]
            # store which file these variables came from
            self.varfiles = [model_varnames[key][0] for key in gvar_list]
            self.gvarfiles = gvar_list
            cdf_data.close()

            # collect all possible variables in set of files and return
            if variables_requested == 'all':
                self.var_dict = {value[0]: value[1:] for key, value in
                                 model_varnames.items() if value[0] in
                                 self.varfiles}
                return

            # store mapping for each variable desired
            self.variables = {model_varnames[gvar][0]: {
                'units': model_varnames[gvar][-1],
                'data': p} for gvar in gvar_list}

            if printfiles:
                print(f'{len(self.filename)} Files:')
                files = self.filename.split(',')
                for f in files:
                    print(f)

            # register interpolators for each variable
            t_reg = perf_counter()
            # store original list b/c gridded interpolators change keys list
            varname_list = [key for key in self.variables.keys()]
            for varname in varname_list:
                self.register_variable(varname, gridded_int)

            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        # define and register the variable
        def register_variable(self, varname, gridded_int):
            """Registers a 3d interpolator with 3d signature"""

            # define and register the interpolators
            key = self.variables[varname]['data']
            coord_dict = {'time': {'units': 'hr', 'data':
                                   self.times[key]['all']},
                          'lon': {'units': 'deg', 'data': self._lon},
                          'lat': {'units': 'deg', 'data': self._lat}}
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]  # variable name in file

            def func(i, fi):  # i = file#, fi = slice#
                cdf_data = Dataset(self.pattern_files[key][i])
                data = array(cdf_data.variables[gvar][fi])
                cdf_data.close()
                return data

            # functionalize the 3D or 4D dataset, series of time slices
            self = RU.Functionalize_Dataset(
                self, coord_dict, varname, self.variables[varname],
                gridded_int, coord_str, interp_flag=3, func=func,
                times_dict=self.times[key])
            return

    return MODEL
